# -*- coding: utf-8 -*-
"""
Created on Sat Feb 04 09:32:46 2017

@author: Zlatko K. Minev, pyEPR ream
"""
from __future__ import division, print_function, absolute_import   # Python 2.7 and 3 compatibility
import platform         # Which OS we run
import numpy as np
import pandas as pd
import warnings
import matplotlib.pyplot as plt

# Constants
from collections import OrderedDict
from ..calcs.constants import Planck, elementary_charge, epsilon_0, pi, π, ħ, ϕ0, e_el


# ==============================================================================
# Utility functions
# ==============================================================================

def combinekw(kw1, kw2):
    ''' Copy kw1,  update with kw2, return result '''
    kw = kw1.copy()
    kw.update(kw2)
    return kw


def isint(value):
    try:
        int(value)
        return True
    except ValueError:
        return False


def isfloat(value):
    try:
        float(value)
        return True
    except ValueError:
        return False


def floor_10(x):
    ''' round to nearest lower power of 10 c'''
    return 10.**(np.floor(np.log10(x)))


def fact(n):
    ''' Factorial '''
    if n <= 1:
        return 1
    return n * fact(n-1)


def nck(n, k):
    ''' choose '''
    return fact(n)/(fact(k)*fact(n-k))


def get_above_diagonal(M):
    ''' extract the values that are above the diagonal.
        Assumes square matrix
    '''
    return M[np.triu_indices(M.shape[0], k=1)]


def df_find_index(s: pd.Series, find, degree=2, ax=False):
    """
    Given a Pandas Series such as of freq with index Lj,
    find the Lj that would give the right frequency
    """
    max_ = max(s.index.values)
    min_ = min(s.index.values)
    if find <= max_ and find >= min_:
        # interpolate
        z = pd.Series(list(s.index.values)+[np.NaN], index=list(s) + [find])
        z = z.sort_index()
        z = z.interpolate()
        return z[find], z
    else:
        print('extrapolating')
        z = pd.Series(list(s.index.values), index=list(s))
        p = df_extrapolate(z, degree=degree, ax=False)
        value = p(find)
        return value, p


def df_interpolate_value(s: pd.Series, find, ax=False, method='index'):
    """
    Given a Pandas Series such as of freq with index Lj,
    find the freq that would correspnd to Lj given a value not in the index
    """
    z = pd.Series(list(s) + [np.NaN], index=list(s.index.values)+[find])
    z = z.sort_index()
    z = z.interpolate(method=method)
    return z[find], z


def df_extrapolate(s, degree=2, ax=False, rng_scale=2.):
    """
    For a pandas series

    Returns np.poly1d
    """
    z = np.polyfit(s.index.values, s.values, degree)
    p = np.poly1d(z)

    if ax:
        if ax is True:
            ax = plt.gca()

        max_ = max(s.index.values)
        min_ = min(s.index.values)
        rng = max_ - min_
        xp = np.linspace(min_-rng_scale*rng, max_+rng_scale*rng, 100)
        ys = p(xp)
        ax.plot(xp, ys)
        s.plot(marker='o', ax=ax)

    return p


def df_regress_value(s: pd.Series, index, degree=2, ax=False, method='index',
                     rng_scale=2.):
    """
    for pandas series.
    calls either  df_interpolate_value or df_extrapolate
    """
    max_ = max(s.index.values)
    min_ = min(s.index.values)
    if index > max_ or index < min_:
        # print('extrapolate')
        p = df_extrapolate(s, degree=degree, ax=ax, rng_scale=rng_scale)
        value = p(index)
    else:
        value = df_interpolate_value(s, index, method=method)[0]
    return value


def series_of_1D_dict_to_multi_df(Uj_ind: pd.Series):
    df = pd.DataFrame(dict([(k, v) for k, v in Uj_ind.items()])).transpose()
    df.index.set_names(Uj_ind.index.names, inplace=True)
    return df


def sort_df_col(df):
    '''         sort by numerical int order    '''
    return df.sort_index(axis=1)

    # Buggy code, deosnt handles ints as inputs or floats as inpts
    col_names = df.columns
    if np.all(col_names.map(isint)):
        return df[col_names.astype(int).sort_values().astype(str)]
    elif np.all(col_names.map(isfloat)):
        # raises error in some cases
        return df[col_names.astype(float).sort_values().astype(str)]
    else:
        return df


def sort_Series_idx(sr):
    '''         sort by numerical int order    '''
    idx_names = sr.index
    if np.all(idx_names.map(isint)):
        return sr[idx_names.astype(int).sort_values().astype(str)]
    if np.all(idx_names.map(isfloat)):
        return sr[idx_names.astype(float).sort_values().astype(str)]
    else:
        return sr


def get_instance_vars(obj, Forbidden=[]):
    VARS = {}
    for v in dir(obj):
        if not (v.startswith('__') or v.startswith('_')):
            if not callable(getattr(obj, v)):
                if not (v in Forbidden):
                    VARS[v] = getattr(obj, v)
    return VARS


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emmitted
    when the function is used. See StackExchange"""
    def newFunc(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn("Call to deprecated function {}.".format(
            func.__name__), category=DeprecationWarning, stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)
    newFunc.__name__ = func.__name__
    newFunc.__doc__ = func.__doc__
    newFunc.__dict__.update(func.__dict__)
    return newFunc


def info_str_platform():
    return '''

 System platform information:

    system   : %s
    node     : %s
    release  : %s
    machine  : %s
    processor: %s
    summary  : %s
    version  : %s

 Python platform information:

    version  : %s (implem: %s)
    compiler : %s

    ''' % (
        platform.system(),
        platform.node(),
        platform.release(),
        platform.machine(),
        platform.processor(),
        platform.platform(),
        platform.version(),
        platform.python_version(), platform.python_implementation(),
        platform.python_compiler())


# ==============================================================================
# Matrix
# ==============================================================================


def print_matrix(M, frmt="{:7.2f}", append_row=""):
    M = np.mat(M)
    for row in np.array(M.tolist()):
        print(' ', end='')
        for chi in row:
            print(frmt.format(chi), end='')
        print(append_row+"\n", end='')


def divide_diagonal_by_2(CHI0, div_fact=2.):
    CHI = CHI0.copy()
    CHI[np.diag_indices_from(CHI)] /= div_fact
    return CHI


def print_NoNewLine(text):
    print((text), end='')


def print_color(text, style=0, fg=24, bg=43, newline=True):
    '''For newer, see pc (or Print_colors)
       style 0..8;   fg  30..38;  bg  40..48
     '''
    format = ';'.join([str(style), str(fg), str(bg)])
    s = '\x1b[%sm %s \x1b[0m' % (format, text)
    if newline:
        print(s)
    else:
        print(s, end='')


class Print_colors:
    '''Colors class:reset all colors with colors.reset; two
    sub classes fg for foreground
    and bg for background; use as colors.subclass.colorname.
    i.e. colors.fg.red or colors.bg.greenalso, the generic bold, disable,
    underline, reverse, strike through,
    and invisible work with the main class i.e. colors.bold
    https://www.geeksforgeeks.org/print-colors-python-terminal/

    Example use:
    ..codeblock python
        print(colors.bg.green, "adgd", colors.fg.red, "dsgdsg")
        print(colors.bg.lightgrey, "dsgsd", colors.fg.red, "sdgsd")
    '''
    reset = '\033[0m'
    bold = '\033[01m'
    disable = '\033[02m'
    underline = '\033[04m'
    reverse = '\033[07m'
    strikethrough = '\033[09m'
    invisible = '\033[08m'

    class fg:
        black = '\033[30m'
        red = '\033[31m'
        green = '\033[32m'
        orange = '\033[33m'
        blue = '\033[34m'
        purple = '\033[35m'
        cyan = '\033[36m'
        lightgrey = '\033[37m'
        darkgrey = '\033[90m'
        lightred = '\033[91m'
        lightgreen = '\033[92m'
        yellow = '\033[93m'
        lightblue = '\033[94m'
        pink = '\033[95m'
        lightcyan = '\033[96m'

    class bg:
        black = '\033[40m'
        red = '\033[41m'
        green = '\033[42m'
        orange = '\033[43m'
        blue = '\033[44m'
        purple = '\033[45m'
        cyan = '\033[46m'
        lightgrey = '\033[47m'


pc = Print_colors

# ==============================================================================
# %%         Dataframe
# ==============================================================================


def DataFrame_col_diff(PS, indx=0):
    ''' check weather the columns of a dataframe are equal,
        returns a T/F series of the row index that specifies which rows are differnt
        USE:
            PS[DataFrame_col_diff(PS)]
    '''
    R = []
    for i in range(PS.shape[1]-1):
        R += [PS.iloc[:, i] == PS.iloc[:, i+1]]
    if len(R) == 1:
        return np.logical_not(R[0])
    else:
        return np.logical_not(np.logical_and.reduce(R))  # pylint: disable=no-member


def DataFrame_display_side_by_side(*args, do_display=True):
    '''
    from pyEPR.toolbox.pythonic import display_dfs
    https://stackoverflow.com/questions/38783027/jupyter-notebook-display-two-pandas-tables-side-by-side
    '''
    from IPython.display import display_html
    html_str = ''
    for df in args:
        html_str += df.to_html()
    text = html_str.replace('table', 'table style="display:inline"')
    if do_display:
        display_html(text, raw=True)
    return text


display_dfs = DataFrame_display_side_by_side


def xarray_unravel_levels(arr, names, my_convert=lambda x: x):
    ''' Takes in nested dict of dict of dataframes
        names : names of lists; you dont have to include the last two dataframe columns & rows, but you can to override them
        requires  xarray
    '''
    import xarray  # pylint: disable=import-error
    if type(arr) == pd.DataFrame:
        return xarray.DataArray(arr, dims=None if len(names) == 0 else names)
    elif type(arr) in [OrderedDict, dict]:
        return xarray.concat([xarray_unravel_levels(item, names[1:]) for k, item in arr.items()], pd.Index(arr.keys(), name=names[0]))
    elif type(arr) == xarray.DataArray:
        return arr
    else:
        return my_convert(arr)


def robust_percentile(calc_data, ROBUST_PERCENTILE=2.):
    '''
        analysis helper function
    '''
    vmin = np.percentile(calc_data, ROBUST_PERCENTILE)
    vmax = np.percentile(calc_data, 100 - ROBUST_PERCENTILE)
    return vmin, vmax


__all__ = ['fact', 'nck', 'combinekw',
           'divide_diagonal_by_2', 'df_find_index',
           'sort_df_col', 'sort_Series_idx',
           'print_matrix', 'print_NoNewLine',
           'DataFrame_col_diff', 'xarray_unravel_levels', 'robust_percentile']



"""
 --- DO NOT MODIFY THIS FILE ---

Default configuration file for pyEPR

This file is NOT meant for users to modify.
Rather, a user should update any config settings they want
in a dictionary called CONFIG in a file called config.py

@author: Zlatko Minev and the pyEPR team
@date: Created on Fri Oct 30 14:21:45 2015
"""

import collections.abc
from . import Dict

# If we are reloading the package, then config will already be defined, then do not overwrite it.
__config_defined__ = 'config' in locals()


config = Dict( # pylint: disable=invalid-name

    # Folder to save result data to.
    root_dir=r'C:\data-pyEPR',
    save_format=r'%Y-%m-%d %H-%M-%S',

    ansys=Dict(
        # method_calc_P_mj sets the method used to calculate the participation ratio in eigenmode.
        # Valud values:
        # 	'line_voltage' : Uses the line voltage integral
        # 	'J_surf_mag'   : takes the avg. Jsurf over the rect. Make sure you have seeded
        # 					lots of tets here. I recommend starting with 4 across smallest dimension.
        # 					Multi-junction calculation of energy participation ratio matrix based on <I_J>.
        # 					Current is integrated average of J_surf by default: (zkm 3/29/16)
        # 					Will calculate the Pj matrix for the selected modes for the given junctions
        # 					junc_rect array & length of junctions
        method_calc_P_mj='line_voltage',

        # To save or not the mesh statistics from an HFSS run
        save_mesh_stats=True,
    ),

    epr = Dict(

        # Define the participation renomalizaiton method
        # False : no extra renormalization to enforce
        #         can be more problematic for large pj, when sim isn't well converged
        # True or 1 : use enforcement of U_J_total to be U_mode-U_H
        #         can be more problematic for small pj, when sim isn't well converged
        # 2     : use enforcement of U_J_total to be U_mode-U_H (i.e., 1)
        #         only when the total particiaption is above a certain threshold
        #         preffered method.
        renorm_pj = 2,
    ),

    # Loss properties of various materials and surfaces
    dissipation=Dict(

        ##################################################
        # Bulk dielectric
        # refs: https://arxiv.org/abs/1308.1743
        #       http://arxiv.org/pdf/1509.01854.pdf
        tan_delta_sapp=1e-6,  # tan(delta) for bulk surface
        epsi=10,    # dielectric

        ##################################################
        # Surface dielectric
        # ref: http://arxiv.org/pdf/1509.01854.pdf

        # Surface dielectric (dirt) thickness
        # units: meters
        th=3e-9,

        # Surface dielectric (dirt) constant
        # units: relative permitivity
        eps_r=10,

        # Surface dielectric (dirt) loss tangent
        # units: unitless, since this is tan(delta)
        tan_delta_surf=1e-3,

        ##################################################
        # Thin-film surface loss
        # units:  Ohms
        # ref:    https://arxiv.org/abs/1308.1743
        surface_Rs=250e-9,

        ##################################################
        # Seam current loss
        # units: per Ohm meter; i.e., seam conductance
        # ref:   http://arxiv.org/pdf/1509.01119.pdf
        gseam=1.0e3,
    ),

    plotting=Dict(
        # Default color map for plottng. Better if made into a string name
        # taken from matplotlib.cm
        default_color_map='viridis',  # pylint: disable=no-member
    ),

    # Not to be used by the user. Just internal
    internal=Dict(

        # Are we using ipython
        ipython=None,

        # Error message for loading packages
        error_msg_missing_import="""\N{face with head-bandage}
        If you need a part of pyEPR that uses this package,
        then please install it. Then add it to the system path (if needed).
        See online setup instructions at
            http://www.github.com/zlatko-minev/pyEPR""",

        # Warn on missing import
        warn_missing_import=False,
    ),

    # Logging
    log=Dict(

        # '%(name)s - %(levelname)s - %(message)s\n   ::%(pathname)s:%(lineno)d: %(funcName)s\n')
        format='%(levelname)s %(asctime)s [%(funcName)s]: %(message)s',

        datefmt='%I:%M%p', #'%I:%M%p %Ss'

        level='INFO'
    )

)


def is_using_ipython():
    """Check if we're in IPython.

    Returns:
        bool -- True if ran in IPython
    """
    try:
        __IPYTHON__  # pylint: disable=undefined-variable, pointless-statement
        return True
    except NameError:
        return False


def update_recursive(d:collections.abc.Mapping, u:collections.abc.Mapping):
    """Recursive update of dictionaries.

    Arguments:
        d {collections.abc.Mapping} -- dict to overwrite
        u {collections.abc.Mapping} -- dcit used to update

    Returns:
        same as d; Updated d
    """
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_recursive(d.get(k, {}), v)
        else:
            d[k] = v
    return d

def get_config():
    """Returns the config pointer.

    If the config is not yet loaded, it will load the defualt config and then
    update it with the _config_user.config dictionary.

    Else, it will just return the pointer to the above-updated config, which the
    user could have modified. The modificaitons will be kept.

    Returns:
        Dict : the config dictionary
    """
    if __config_defined__:
        #print('Config is already defined.') # not sure we ever make it here
        return config

    else:
        # Config is only loaded for the first time, set it up.
        #print('First time load of config')

        # Update with user config
        from . import _config_user
        _config = update_recursive(config, _config_user.config)

        # Add to config any bootup params
        config.internal.ipython = is_using_ipython()

        return config


__all__ = ['get_config']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 19 18:14:08 2019

Unit and variable conversions.

@author: Zlatko Minev
"""

from __future__ import (absolute_import,  # Python 2.7 and 3 compatibility
                        division, print_function)

import numpy as np
import pandas as pd
from numpy import sqrt

from .basic import CalcsBasic
from .constants import (Planck, e_el, elementary_charge, fluxQ, hbar, pi, ħ, π,
                        ϕ0)


class Convert():
    '''
        Static container class for conversions of units and variables.

        TEST CONVERSION:
        ```python
            from pyEPR.toolbox.conversions import Convert

            Lj_nH, Cs_fF = 11, 60
            Convert.transmon_print_all_params(Lj_nH, Cs_fF);

        ```
    '''
    # Known SI prefixed
    _prefix = {'y': -24,  # yocto
               'z': -21,  # zepto
               'a': -18,  # atto
               'f': -15,  # femto
               'p': -12,  # pico
               'n': -9,   # nano
               'u': -6,   # micro
               'm': -3,   # mili
               'c': -2,   # centi
               'd': -1,   # deci
               ' ': 0,
               'k': 3,    # kilo
               'M': 6,    # mega
               'G': 9,    # giga
               'T': 12,   # tera
               'P': 15,   # peta
               'E': 18,   # exa
               'Z': 21,   # zetta
               'Y': 24,   # yotta
               }

    # Known SI units
    _SI_units = ['H',  # Henries
                 'F',  # Farads
                 'Hz',  # Hertz
                 'Ohm',  # Ohms
                 'Ω',  # Ohms
                 'Wb'  # Webers
                 'J',  # Joules
                 'A'   # Amps
                 ]

    @staticmethod
    def toSI(number, from_units: str):
        r"""
        Convert from SI unit prefix to regular SI units
        If the from_units is ' ' or not in the prefix list,
        then the unit is assumed to be
        """
        if from_units in Convert._SI_units:
            from_units = ' '
        # else: we assume that the first letter is a prefix
        return number*(10**Convert._prefix.get(from_units[0]))

    @staticmethod
    def fromSI(number, from_units: str):
        r"""Convert a number with SI units, such as fF to F.

        Arguments:
            number {[numeric]} -- number
            from_units {str} -- string

        Returns:
            numeric number, with units expanded
        """
        if from_units in Convert._SI_units:
            from_units = ' '
        # else: we assume that the first letter is a prefix
        return number*(10**(-Convert._prefix.get(from_units[0])))

    @staticmethod
    def _convert_num(out_func, in_num, in_units, out_units):
        in_num = 1.0*in_num  # to float
        # convert units of input number
        in_num = Convert.toSI(in_num, in_units)
        out_num = out_func(in_num)  # Assume func processes all in SI units
        out_num = Convert.fromSI(out_num, out_units)
        return out_num

    @staticmethod
    def Ej_from_Lj(Lj, units_in='nH', units_out='MHz'):
        r'''
        Josephson Junction energy from Josephson inductance.
        Returns in MHz

        $E_j = \phi_0^2 / L_J$
        '''
        return Convert._convert_num(
            # Plank to go from Joules to Hz
            lambda _Lj: Planck**-1 * (ϕ0**2)/_Lj,
            Lj, units_in, units_out)

    @staticmethod
    def Lj_from_Ej(Ej, units_in='MHz', units_out='nH'):
        r'''
        Josephson Junction ind from Josephson energy in MHZ.
        Returns in units of nano Henries by default

        $E_j = \phi_0^2 / L_J$
        '''
        return Convert._convert_num(
            lambda _x: (ϕ0**2.)/(_x*Planck),  # Plank to go from Joules to Hz
            Ej, units_in, units_out)

    @staticmethod
    def Ic_from_Lj(Lj, units_in='nH', units_out='nA'):
        r'''
        Josephson Junction crit. curr from Josephson inductance.

        $E_j = \phi_0^2 / L_J = \phi_0 I_C $
        '''
        return Convert._convert_num(
            lambda _x: ϕ0/_x,  # Plank to go from Joules to Hz
            Lj, units_in, units_out)

    @staticmethod
    def Lj_from_Ic(Lj, units_in='nA', units_out='nH'):
        r'''
        Josephson Junction crit. curr from Josephson inductance.

        $E_j = \phi_0^2 / L_J = \phi_0 I_C $
        '''
        return Convert._convert_num(
            lambda _x: ϕ0/_x,  # Plank to go from Joules to Hz
            Lj, units_in, units_out)

    @staticmethod
    def Ec_from_Cs(Cs,  units_in='fF', units_out='MHz'):
        r'''
        Charging energy 4Ec n^2, where n=Q/2e
        Returns in MHz

        $E_{C}=\frac{e^{2}}{2C}J$
        '''
        return Convert._convert_num(
            # Plank to go from Joules to Hz
            lambda _x: Planck**-1 * (e_el**2.)/(2.*_x),
            Cs, units_in, units_out)

    @staticmethod
    def Cs_from_Ec(Ec, units_in='MHz', units_out='fF'):
        r'''
        Charging energy 4Ec n^2, where n=Q/2e

        Returns in SI units, in Farads.

        $E_{C}=\frac{e^{2}}{2C}J$
        '''
        return Convert._convert_num(
            # Plank to go from Joules to Hz
            lambda _x: (e_el**2.)/(2.*_x*Planck),
            Ec, units_in, units_out)

    @staticmethod
    def ZPF_from_LC(L, C):
        r'''
        Input units assumed to be identical

        Returns Phi ZPF in and Q_ZPF in NOT reduced units, but SI
        '''
        Z = sqrt(L/C)
        return (sqrt(hbar*Z/2.), sqrt(hbar/(2.*Z)))  # Phi , Q

    @staticmethod
    def ZPF_from_EPR(hfss_freqs, hfss_epr_, hfss_signs, hfss_Ljs,
                     Lj_units_in='H', to_df=False):
        r"""
        Parameters:
            Can be either Pandas or numpy arrays.

            hfss_freqs : HFSS Freqs. (standard units: GHz, but these will cancel with Ejs) (list/Series)
            hfss_epr : EPR ratio matrix, dim = M x J (2D array/DataFrame)
            hfss_signs : Sign matrix, dim = M x J  (2D array/DataFrame)
            hfss_Ljs : Assumed in Henries (see Lj_units_in). (list/Series)

            Lj_units_in : Default 'H' for Henries. Can change here.

        Returns:
            M x J matrix of reduced ZPF; i.e., scaled by reduced flux quantum.
            type: np.array
            and a tuple of matricies.

        Example use:
            ϕzpf, (Ωm, Ej, Pmj, Smj) = Convert.ZPF_from_EPR(hfss_freqs, hfss_epr, hfss_signs, hfss_Ljs, to_df=True)
        """

        hfss_freqs, hfss_epr, hfss_signs, hfss_Ljs = map(
            np.array, (hfss_freqs, hfss_epr_, hfss_signs, hfss_Ljs))

        Ωd = np.diagflat(hfss_freqs)
        Ej = Convert.Ej_from_Lj(
            hfss_Ljs, units_in=Lj_units_in, units_out='GHz')
        Ej = np.diagflat(Ej)

        ϕzpfs = CalcsBasic.epr_to_zpf(hfss_epr, hfss_signs, Ωd, Ej)

        if to_df:
            ϕzpfs = pd.DataFrame(
                ϕzpfs, columns='ϕ'+hfss_epr_.columns.values, index=hfss_epr_.index)

        return ϕzpfs, (Ωd, Ej, hfss_epr, hfss_signs)

    @staticmethod
    def Omega_from_LC(L, C):
        r'''
        Calculate the resonant *angular* frequency
        '''
        return sqrt(1./(L*C))



"""
Module for reporting utility functions

@author: Zlatko K Minev
"""

import pandas as pd

from .toolbox.plotting import legend_translucent, plt


def _style_plot_convergence(ax, ylabel=None, xlabel='Pass number', ylabel_col='k', y_title=False):
    ax.set_xlabel(xlabel)
    if ylabel:
        if y_title:
            ax.set_title(ylabel)
        else:
            ax.set_ylabel(ylabel, color=ylabel_col)
    ax.grid()
    ax.autoscale(tight=False)
    ax.set_axisbelow(True)  # Don't allow the axis to be on top of your data
    ax.minorticks_on()
    ax.grid(which='minor', linestyle=':',
            linewidth='0.5', color='black', alpha=0.2)
    ax.grid(which='major', alpha=0.5)


_style_plot_conv_kw = dict(marker='o', ms=4)


def plot_convergence_max_df(ax, s, kw={}, color='r'):
    '''For a single pass'''
    s.plot(ax=ax, **{**dict(c='r'), **_style_plot_conv_kw, **kw})
    ax.set_yscale("log")
    _style_plot_convergence(ax)
    fig = ax.figure
    fig.text(0.45, 0.95, s.name, ha="center",
             va="bottom", size="medium", color=color)
    ax.tick_params(axis='y', labelcolor=color)
    #ax.axhline(1.0, color='k', lw=1.5,alpha= 0.35)
    #ax.axhline(0.1, color='k', lw=1.5,alpha= 0.35)
    ax.grid(which='minor', linestyle=':',
            linewidth='0.5', color=color, alpha=0.25)
    ax.grid(which='major', color='#c4abab', alpha=0.5)
    ax.spines['left'].set_color(color)


def plot_convergence_solved_elem(ax, s, kw={}, color='b'):
    '''For a single pass'''
    (s/1000).plot(ax=ax, **{**dict(c='b'), **_style_plot_conv_kw, **kw})
    _style_plot_convergence(ax)
    # ax.set_ylim([100,None])
    # ax.set_yscale("log")
    ax.minorticks_off()
    ax.grid(False)
    ax.tick_params(axis='y', labelcolor=color)
    # ax.ticklabel_format(style='sci',scilimits=(0,0))
    fig = ax.figure
    fig.text(0.6, 0.95, 'Solved elements (1000s)', ha="center",
             va="bottom", size="medium", color=color)
    ax.spines['left'].set_color('r')
    ax.spines['right'].set_color(color)


def plot_convergence_f_vspass(ax, s, kw={}):
    '''For a single pass'''
    if s is not None:
        (s).plot(ax=ax, **{**_style_plot_conv_kw, **kw})
        _style_plot_convergence(ax, 'Eigenmode f vs. pass [GHz]', y_title=True)
        legend_translucent(ax, leg_kw=dict(fontsize=6))


def plot_convergence_maxdf_vs_sol(ax, s, s2, kw={}):
    '''
    ax, 'Max Δf %', 'Solved elements', kw for plot
    '''
    s = s.copy()
    s.index = s2
    (s).plot(ax=ax, **{**_style_plot_conv_kw, **kw})
    _style_plot_convergence(ax, s.name, xlabel='Solved elements', y_title=True)
    ax.set_yscale("log")
    ax.set_xscale("log")


# quick and dirty use
def _plot_q3d_convergence_main(epr, RES):
    fig = epr.hfss_report_full_convergence(_display=False)

    ax = fig.axes[0]
    ax2 = ax.twinx()
    ax.cla()
    ax2.cla()
    RES['alpha'].plot(ax=ax, c='b')
    (RES['fQ']*1000).plot(ax=ax2, c='red')
    from matplotlib import pyplot as plt
    _style_plot_convergence(
        ax, 'Alpha (blue),  Freq (red) [MHz]', y_title=True)
    ax2.set_ylabel('Frequency (MHz)', color='r')
    ax.set_ylabel('Alpha(MHz)', color='b')
    ax2.spines['right'].set_color('r')
    ax2.tick_params(axis='y', labelcolor='r')
    ax.tick_params(axis='y', labelcolor='b')
    # legend_translucent(ax)
    # legend_translucent(ax2)
    ax.set_xlabel('Pass')
    fig.tight_layout()

    return fig


def _plot_q3d_convergence_chi_f(RES):
    df_chi = pd.DataFrame(RES['chi_in_MHz'].values.tolist())
    df_chi.index.name = 'Pass'
    df_g = pd.DataFrame(RES['gbus'].values.tolist())
    df_g.index.name = 'Pass'

    fig, axs = plt.subplots(1, 2, figsize=(9, 3.5))
    df_chi.plot(lw=2, ax=axs[0])
    df_g.plot(lw=2, ax=axs[1])
    _style_plot_convergence(axs[0])
    _style_plot_convergence(axs[1])
    axs[0].set_title(r'$\chi$ convergence (MHz)')
    axs[1].set_title(r'$g$ convergence (MHz)')

    return fig



"""
pyEPR constants and convinience definitions.

@author: Zlatko Minev
"""
# pylint: disable=invalid-name

from scipy.constants import Planck, elementary_charge, epsilon_0, pi  # pylint: disable=unused-import

# Pi
π = pi

# Reduced Planks constant
ħ = hbar = Planck/(2*pi)

# Reduced Flux Quantum  (3.29105976 × 10-16 Webers)
ϕ0 = fluxQ = ħ / (2*elementary_charge)

# Magnitude of the electric charge carried by a single electron
e_el = elementary_charge



"""
Main interface module to use pyEPR.

Contains code that works on the analysis after hfss, ansys, etc. These can now be closed.

Copyright Zlatko Minev, Zaki Leghtas, and the pyEPR team
2015, 2016, 2017, 2018, 2019, 2020
"""
# pylint: disable=invalid-name
# todo remove this pylint hack later

from __future__ import print_function  # Python 2.7 and 3 compatibility

from typing import List
import pickle
import sys
import time
from collections import OrderedDict
from pathlib import Path
from .calcs.convert import Convert

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from IPython.display import Markdown, display
from numpy.linalg import inv

# pyEPR custom imports
from . import Dict, config, logger
from .ansys import ureg
from .calcs.back_box_numeric import epr_numerical_diagonalization
from .calcs.basic import CalcsBasic
from .calcs.constants import Planck, fluxQ
from .core_distributed_analysis import DistributedAnalysis
from .toolbox.plotting import cmap_discrete, legend_translucent
from .toolbox.pythonic import (DataFrame_col_diff, divide_diagonal_by_2,
                               print_color, print_matrix, sort_df_col,
                               sort_Series_idx, df_find_index, series_of_1D_dict_to_multi_df)

from .reports import (plot_convergence_max_df, plot_convergence_solved_elem)

class HamiltonianResultsContainer(OrderedDict):
    """
    The user should only use the QuantumAnalysis class interface.

    This class is largely for internal use.

    It is a dictionary based class to contain the results stored.
    """

    file_name_extra = ' HamiltonianResultsContainer.npz'

    def __init__(self, dict_file=None, data_dir=None):
        """ input:
           dict file - 1. ethier None to create an empty results hamilitoninan as
                       as was done in the original code

                       2. or a string with the name of the file where the file of the
                       previously saved HamiltonianResultsContainer instatnce we wish
                       to load

                       3. or an existing instance of a dict class which will be
                       upgraded to the HamiltonianResultsContainer class

            data_dir -  the directory in which the file is to be saved or loaded
                        from, defults to the config.root_dir
        """

        super().__init__()

        self.sort_index = True  # for retrieval

        if data_dir is None:
            data_dir = Path(config.root_dir) / 'temp' / \
                time.strftime('%Y-%m-%d %H-%M-%S', time.localtime())

        data_dir = Path(data_dir).resolve()
        file_name = data_dir.stem
        directory = data_dir.parents[0]
        if not directory.is_dir():
            directory.mkdir(parents=True, exist_ok=True)

        if dict_file is None:
            self.file_name = str(
                directory/(str(file_name)+self.file_name_extra))
            #logger.info(f'Filename hamiltonian params to {self.file_name }')

        elif isinstance(dict_file, str):
            try:
                self.file_name = str(data_dir)+'\\' + dict_file
                self.load()
            except:
                self.file_name = dict_file
                self.load()

        elif isinstance(dict_file, dict):
            # Depreciated
            self._inject_dic(dict_file)
            self.file_name = str(data_dir)+self.file_name_extra

        else:
            raise ValueError(
                'type dict_file is of type {}'.format(type(dict_file)))
            # load file

    def save(self, filename: str = None):
        """
        Uses numpy npz file.
        """

        if filename is None:
            filename = self.file_name

        np.savez(filename, Res_Hamil=dict(self))
        return filename

    def load(self, filename=None):
        """
        Uses numpy npz file.
        """
        if filename is None:
            filename = self.file_name

        self._inject_dic(extract_dic(file_name=filename)[0])
        return filename

    def _inject_dic(self, add_dic):
        Init_number_of_keys = len(self.keys())
        for key, val in add_dic.items():
            # TODO remove all copies of same data
            #  if key in self.keys():
                #raise ValueError('trying to overwrite an exsiting varation')
            self[str(int(key)+Init_number_of_keys)] = val
        return 1

    @staticmethod
    def _do_sort_index(z: pd.DataFrame):
        """Overwrite to sort by custom function

        Arguments:
            z {pd.DataFrame} -- Input

        Returns:
            Sorted DtaaFrame
        """
        if isinstance(z, pd.DataFrame):
            return z.sort_index(axis=1)
        else:
            return z

    def vs_variations(self,
                      quantity: str,
                      variations: list = None,
                      vs='variation',
                      to_dataframe=False):
        """

        QUANTITIES:
            `f_0`  : HFSS Frequencies
            `f_1`  : Analyutical first order PT on the p=4 term of the cosine
            `f_ND` : Numerically diagonalized
            `chi_O1`: chi matrix from 1st order PT

        Arguments:
            quantity {[type]} -- [description]

        Keyword Arguments:
            variations {list of strings} -- Variations (default: {None} -- means all)
            vs {str} -- Swept against (default: {'variation'})
            to_dataframe {bool} -- convert or not the result to dataframe.
                         Make sure to call only if it can be converted to a DataFrame or can
                         be concatinated into a multi-index DataFrame

        Returns:
            [type] -- [description]
        """
        variations = variations or self.keys()

        res = OrderedDict()
        for key in variations:
            if vs == 'variation':
                res[key] = self[key][quantity]
            else:
                # convert the key to numeric if possible
                key_new = ureg.Quantity(
                    self[key]['hfss_variables']['_'+vs]).magnitude
                res[key_new] = self[key][quantity]

        # Convert to dataframe
        z = res
        if to_dataframe:  # only call if z can be converted to a dataframe
            z = sort_df_col(pd.DataFrame(z))
            if self.sort_index:
                z = self._do_sort_index(z)
            z.index.name = 'eigenmode'
            z.columns.name = vs

        return z

    # Quick lookup function

    def get_frequencies_HFSS(self, variations: list = None, vs='variation'):
        '''See help for `vs_variations`'''
        return self.vs_variations('f_0', variations=variations, vs=vs, to_dataframe=True)

    def get_frequencies_O1(self, variations: list = None, vs='variation'):
        '''See help for `vs_variations`'''
        return self.vs_variations('f_1', variations=variations, vs=vs, to_dataframe=True)

    def get_frequencies_ND(self, variations: list = None, vs='variation'):
        '''See help for `vs_variations`'''
        return self.vs_variations('f_ND', variations=variations, vs=vs, to_dataframe=True)

    def get_chi_O1(self, variations: list = None, vs='variation'):
        return self.vs_variations('chi_O1', variations=variations, vs=vs)

    def get_chi_ND(self, variations: list = None, vs='variation'):
        return self.vs_variations('chi_ND', variations=variations, vs=vs)


class QuantumAnalysis(object):
    '''
    Defines an analysis object which loads and plots data from a h5 file
    This data is obtained using DistributedAnalysis
    '''

    def __init__(self, data_filename,
                 variations: list = None,
                 do_print_info=True,
                 Res_hamil_filename=None):

        self.data_filename = data_filename
        self.results = HamiltonianResultsContainer(dict_file=Res_hamil_filename,
                                                   data_dir=data_filename)

        with open(str(data_filename), 'rb') as handle:
            # Contain everything: project_info and results
            self.data = Dict(pickle.load(handle))

        # Reverse from variations on outside to on inside
        results = DistributedAnalysis.results_variations_on_inside(
            self.data.results)

        # Convinience functions
        self.variations = variations or list(self.data.results.keys())
        self._hfss_variables = results['hfss_variables']
        self.freqs_hfss = results['freqs_hfss_GHz']
        self.Qs = results['Qs']
        self.Qm_coupling = results['Qm_coupling']
        self.Ljs = results['Ljs']  # DataFrame
        self.Cjs = results['Cjs']  # DataFrame
        self.OM = results['Om']  # dict of dataframes
        self.PM = results['Pm']  # participation matrices - raw, unnormed here
        # participation matrices for capactive elements
        self.PM_cap = results['Pm_cap']
        self.SM = results['Sm']  # sign matrices
        self.I_peak = results['I_peak']
        self.V_peak = results['V_peak']
        self.modes = results['modes']

        self.sols = results['sols']
        self.ansys_energies = results.get('ansys_energies', {})

        self.mesh_stats = results['mesh']
        self.convergence = results['convergence']
        self.convergence_f_pass = results['convergence_f_pass']

        self.n_modes = len(self.modes['0'])
        self._renorm_pj = config.epr.renorm_pj

        # Unique variation params -- make a get function
        dum = DataFrame_col_diff(self._hfss_variables)
        self.hfss_vars_diff_idx = dum if not (dum.any() == False) else []
        try:
            self.Num_hfss_vars_diff_idx = len(
                self.hfss_vars_diff_idx[self.hfss_vars_diff_idx == True])
        except:
            e = sys.exc_info()[0]
            logger.warning("<p>Error: %s</p>" % e)
            self.Num_hfss_vars_diff_idx = 0

        if do_print_info:
            self.print_info()

    @property
    def project_info(self):
        return self.data.project_info

    def print_info(self):
        print("\t Differences in variations:")
        if len(self.hfss_vars_diff_idx) > 0:
            display(self._hfss_variables[self.hfss_vars_diff_idx])
        print('\n')

    def get_vs_variable(self, swp_var, attr: str):
        """
        Convert the index of a dicitoanry that is stored here from
        variation number to variable value.

        Args:
            swp_var (str) :name of sweep variable in ansys
            attr: name of local attribute, eg.., 'ansys_energies'
        """
        #from collections import OrderedDict
        variable = self.get_variable_vs(swp_var)
        return OrderedDict([(variable[variation], val)
                            for variation, val in getattr(self, attr).items()])

    def get_variable_vs(self, swpvar, lv=None):
        """ lv is list of variations (example ['0', '1']), if None it takes all variations
            swpvar is the variable by which to orginize

            return:
            ordered dicitonary of key which is the variation number and the magnitude
            of swaver as the item
        """
        ret = OrderedDict()
        if lv is None:
            for key, varz in self._hfss_variables.items():
                ret[key] = ureg.Quantity(varz['_'+swpvar]).magnitude
        else:
            try:
                for key in lv:
                    ret[key] = ureg.Quantity(
                        self._hfss_variables[key]['_'+swpvar]).magnitude
            except:
                print(' No such variation as ' + key)
        return ret

    def get_variable_value(self, swpvar, lv=None):

        var = self.get_variable_vs(swpvar, lv=lv)
        return [var[key] for key in var.keys()]

    def get_variations_of_variable_value(self, swpvar, value, lv=None):
        """A function to return all the variations in which one of the variables
            has a specific value lv is list of variations (example ['0', '1']),
            if None it takes all variations
            swpvar is a string and the name of the variable we wish to filter
            value is the value of swapvr in which we are intrested

            returns lv - a list of the variations for which swavr==value
            """

        if lv is None:
            lv = self.variations

        ret = self.get_variable_vs(swpvar, lv=lv)

        lv = np.array(list(ret.keys()))[np.array(list(ret.values())) == value]

        #lv = lv_temp if not len(lv_temp) else lv
        if not (len(lv)):
            raise ValueError('No variations have the variable-' + swpvar +
                             '= {}'.format(value))

        return list(lv)

    def get_variation_of_multiple_variables_value(self, Var_dic, lv=None):
        """
                SEE get_variations_of_variable_value
            A function to return all the variations in which one of the variables has a specific value
            lv is list of variations (example ['0', '1']), if None it takes all variations
            Var_dic is a dic with the name of the variable as key and the value to filter as item
            """

        if lv is None:
            lv = self.variations

        var_str = None
        for key, var in Var_dic.items():
            lv = self.get_variations_of_variable_value(key, var, lv)
            if var_str is None:
                var_str = key + '= {}'.format(var)
            else:
                var_str = var_str + ' & ' + key + '= {}'.format(var)
        return lv, var_str

    def get_convergences_max_tets(self):
        ''' Index([u'Pass Number', u'Solved Elements', u'Max Delta Freq. %' ])  '''
        ret = OrderedDict()
        for key, df in self.convergence.items():
            ret[key] = df['Solved Elements'].iloc[-1]
        return ret

    def get_convergences_tets_vs_pass(self, as_dataframe=True):
        ''' Index([u'Pass Number', u'Solved Elements', u'Max Delta Freq. %' ])  '''
        ret = OrderedDict()
        for key, df in self.convergence.items():
            s = df['Solved Elements']
            s = s.reset_index().dropna().set_index('Pass Number')
            #s.index = df['Pass Number']
            ret[key] = s

        if as_dataframe:
            ret = pd.concat(ret)
            ret = ret.unstack(0)['Solved Elements']

        return ret

    def get_convergences_max_delta_freq_vs_pass(self, as_dataframe=True):
        ''' Index([u'Pass Number', u'Solved Elements', u'Max Delta Freq. %' ])  '''
        KEY = 'Max Delta Freq. %'

        ret = OrderedDict()
        for key, df in self.convergence.items():
            s = df[KEY]
            s = s.reset_index().dropna().set_index('Pass Number')
            #s.index = df['Pass Number']
            ret[key] = s

        if as_dataframe:
            ret = pd.concat(ret)
            ret = ret.unstack(0)[KEY]

        return ret

    def get_mesh_tot(self):
        ret = OrderedDict()
        for key, m in self.mesh_stats.items():
            ret[key] = m['Num Tets  '].sum()
        return ret

    def get_Ejs(self, variation):
        ''' EJs in GHz
        See calcs.convert
        '''
        Ljs = self.Ljs[variation]
        Ejs = fluxQ**2/Ljs/Planck*10**-9
        return Ejs

    def get_Ecs(self, variation):
        ''' ECs in GHz
        Returns as padnas series
        '''
        Cs = self.Cjs[variation]
        return Convert.Ec_from_Cs(Cs,  units_in='F', units_out='GHz')

    def analyze_all_variations(self,
                               variations: List[str] = None,
                               analyze_previous=False,
                               **kwargs):
        '''
        See analyze_variation for full documentation

        Specific params:
        --------------------
            variations : None returns all_variations otherwis this is a list with number
                         as strings ['0', '1']
            nalyze_previous :set to true if you wish to overwrite previous analysis
        '''

        result = OrderedDict()

        if variations is None:
            variations = self.variations

        for variation in variations:
            if (not analyze_previous) and (variation in self.results.keys()):
                result[variation] = self.results[variation]
            else:
                result[variation] = self.analyze_variation(variation, **kwargs)

        self.results.save()

        return result

    def _get_ansys_total_energies(self, variation):
        res = {}
        for getkey in ['U_tot_cap', 'U_tot_ind', 'U_H', 'U_E', 'U_norm']:
            res[getkey] = pd.Series({mode: self.ansys_energies[variation][mode][getkey]
                                     for mode in self.ansys_energies[variation]})
        df = pd.DataFrame(res)
        df.index.name = 'modes'
        return df

    def _get_participation_normalized(self, variation, _renorm_pj=None, print_=False):
        '''
            Get normalized Pmj Matrix

            Return DataFrame object for PJ
        '''
        if _renorm_pj is None:
            _renorm_pj = self._renorm_pj

        # Columns are junctions; rows are modes
        Pm = self.PM[variation].copy()   # EPR matrix DataFrame

        # EPR matrix for capacitor DataFrame
        Pm_cap = self.PM_cap[variation].copy()

        if _renorm_pj:  # just non False

            # Renormalize
            # Should we still do this when Pm_glb_sum is very small
            #s = self.sols[variation]
            # sum of participation energies as calculated by global UH and UE
            # U_mode = s['U_E'] # peak mode energy; or U bar as i denote it sometimes
            # We need to add the capactiro here, and maybe take the mean of that

            energies = self._get_ansys_total_energies(variation)

            U_mode = (energies['U_tot_cap'] + energies['U_tot_ind'])/2.
            U_diff = abs(energies['U_tot_cap'] - energies['U_tot_ind'])/U_mode
            if np.any(U_diff > 0.15):
                logger.error(f"WARNING: U_tot_cap-U_tot_ind / mean = {np.max(np.abs(U_diff))*100:.1f}% is > 15%. \
                    \nIs the simulation converged? Proceed with caution")

            # global sums of participations
            Pm_glb_sum = abs((U_mode-energies['U_H'])/U_mode)
            Pm_cap_glb_sum = abs((U_mode-energies['U_E'])/U_mode)

            # norms
            Pm_norm = Pm_glb_sum/Pm.sum(axis=1)
            Pm_cap_norm = Pm_cap_glb_sum/Pm_cap.sum(axis=1)
            # this is not the correct scaling yet! WARNING. Factors of 2 laying around too
            # these numbers are a bit all over the place for now. very small

            if _renorm_pj == True or _renorm_pj == 1:
                idx = Pm > -1E6  # everywhere scale
                idx_cap = Pm_cap > -1E6
            elif _renorm_pj == 2:
                idx = Pm > 0.15  # Mask for where to scale
                idx_cap = Pm_cap > 0.15
            else:
                raise NotImplementedError(
                    "Unkown _renorm_pj argument or config values!")

            if print_:
                # \nPm_cap_norm=\n{Pm_cap_norm}")
                print(f"Pm_norm=\n{Pm_norm}\n")
                print(f"Pm_norm idx =\n{idx}")


            Pm[idx] = Pm[idx].mul(Pm_norm, axis=0)
            Pm_cap[idx_cap] = Pm_cap[idx_cap].mul(Pm_cap_norm, axis=0)
            #Pm = Pm.mul(Pm_norm, axis=0)
            #Pm_cap = Pm_cap.mul(Pm_cap_norm, axis=0)

        else:
            Pm_norm = 1
            Pm_cap_norm = 1
            idx = None
            idx_cap = None
            if print_:
                print('NO renorm!')

        if np.any(Pm < 0.0):
            print_color("  ! Warning:  Some p_mj was found <= 0. This is probably a numerical error,'\
                'or a super low-Q mode.  We will take the abs value.  Otherwise, rerun with more precision,'\
                'inspect, and do due dilligence.)")
            print(Pm, '\n')
            Pm = np.abs(Pm)

        return {'PJ': Pm, 'Pm_norm': Pm_norm, 'PJ_cap': Pm_cap,
                'Pm_cap_norm': Pm_cap_norm,
                'idx': idx,
                'idx_cap': idx_cap}

    def get_epr_base_matrices(self, variation, _renorm_pj=None, print_=False):
        r'''
        Return the key matricies used in the EPR method for analytic calcualtions.

        All as matrices
            :PJ: Participatuion matrix, p_mj
            :SJ: Sign matrix, s_mj
            :Om: Omega_mm matrix (in GHz) (\hbar = 1) Not radians.
            :EJ: E_jj matrix of Josephson energies (in same units as hbar omega matrix)
            :PHI_zpf: ZPFs in units of \phi_0 reduced flux quantum
            :PJ_cap: capactive particiaption matrix

            Return all as *np.array*
                PM, SIGN, Om, EJ, Phi_ZPF
        '''
        # TODO: superseed by Convert.ZPF_from_EPR

        res = self._get_participation_normalized(
            variation, _renorm_pj=_renorm_pj, print_=print_)

        PJ = np.array(res['PJ'])
        PJ_cap = np.array(res['PJ_cap'])

        # Sign bits
        SJ = np.array(self.SM[variation])                # DataFrame
        #  Frequencies of HFSS linear modes.
        #  Input in dataframe but of one line. Output nd array
        Om = np.diagflat(self.OM[variation].values)      # GHz
        # Junction energies
        EJ = np.diagflat(self.get_Ejs(variation).values)  # GHz
        Ec = np.diagflat(self.get_Ecs(variation).values)  # GHz

        for x in ("PJ", "SJ", "Om", "EJ"):
            logger.debug(f"{x}=")
            logger.debug(locals()[x])

        PHI_zpf = CalcsBasic.epr_to_zpf(PJ, SJ, Om, EJ)
        n_zpf = CalcsBasic.epr_cap_to_nzpf(PJ, SJ, Om, Ec)

        return PJ, SJ, Om, EJ, PHI_zpf, PJ_cap, n_zpf                 # All as np.array

    def analyze_variation(self,
                          variation: List[str],
                          cos_trunc: int = None,
                          fock_trunc: int = None,
                          print_result: bool = True,
                          junctions: List = None,
                          modes: List = None):
        # TODO avoide analyzing a previously analyzed variation
        '''
        Core analysis function to call!

        Args:
        ---------------
            junctions: list or slice of junctions to include in the analysis.
                None defaults to analysing all junctions
            modes: list or slice of modes to include in the analysis.
                None defaults to analysing all modes

        Returns:
        ----------------
            f_0 [MHz]    : Eigenmode frequencies computed by HFSS; i.e., linear freq returned in GHz
            f_1 [MHz]    : Dressed mode frequencies (by the non-linearity; e.g., Lamb shift, etc. ).
                           If numerical diagonalizaiton is run, then we return the numerically diagonalizaed
                           frequencies, otherwise, use 1st order pertuirbation theory on the 4th order
                           expansion of the cosine.
            f_ND [MHz]   : Numerical diagonalizaiton
            chi_O1 [MHz] : Analytic expression for the chis based on a cos trunc to 4th order, and using 1st
                           order perturbation theory. Diag is anharmonicity, off diag is full cross-Kerr.
            chi_ND [MHz] : Numerically diagonalized chi matrix. Diag is anharmonicity, off diag is full
                           cross-Kerr.
        '''

        # ensuring proper matrix dimensionality when slicing
        junctions = (junctions,) if type(junctions) is int else junctions

        if modes is None:
            modes = list(range(self.n_modes))
        
        tmp_n_modes = self.n_modes
        tmp_modes =self.modes[variation]
        self.n_modes = len(modes)
        self.modes[variation]= modes
    

        if (fock_trunc is None) or (cos_trunc is None):
            fock_trunc = cos_trunc = None

        if print_result:
            print('\n', '. '*40)
            print('Variation %s\n' % variation)
        else:
            print('%s, ' % variation, end='')

        # Get matrices
        PJ, SJ, Om, EJ, PHI_zpf, PJ_cap, n_zpf = self.get_epr_base_matrices(
            variation)
        freqs_hfss = self.freqs_hfss[variation].values[(modes)]
        Ljs = self.Ljs[variation].values

        # reduce matrices to only include certain modes/junctions
        if junctions is not None:
            Ljs = Ljs[junctions, ]
            PJ = PJ[:, junctions]
            SJ = SJ[:, junctions]
            EJ = EJ[:, junctions][junctions, :]
            PHI_zpf = PHI_zpf[:, junctions]
            PJ_cap = PJ_cap[:, junctions]

        if modes is not None:
            freqs_hfss = freqs_hfss[range(len(self.modes[variation])), ]
            PJ = PJ[modes, :]
            SJ = SJ[modes, :]
            Om = Om[modes, :][:, modes]
            PHI_zpf = PHI_zpf[modes, :]
            PJ_cap = PJ_cap[:, junctions]

        # Analytic 4-th order
        CHI_O1 = 0.25 * Om @ PJ @ inv(EJ) @ PJ.T @ Om * 1000.  # MHz
        f1s = np.diag(Om) - 0.5*np.ndarray.flatten(np.array(CHI_O1.sum(1))) / \
            1000.                  # 1st order PT expect freq to be dressed down by alpha
        CHI_O1 = divide_diagonal_by_2(CHI_O1)   # Make the diagonals alpha

        # Numerical diag
        if cos_trunc is not None:
            f1_ND, CHI_ND = epr_numerical_diagonalization(freqs_hfss,
                                                          Ljs,
                                                          PHI_zpf,
                                                          cos_trunc=cos_trunc,
                                                          fock_trunc=fock_trunc)
        else:
            f1_ND, CHI_ND = None, None

        result = OrderedDict()
        result['f_0'] = self.freqs_hfss[variation][modes] * \
            1E3  # MHz - obtained directly from HFSS
        result['f_1'] = pd.Series(f1s)*1E3     # MHz
        result['f_ND'] = pd.Series(f1_ND)*1E-6  # MHz
        result['chi_O1'] = pd.DataFrame(CHI_O1)
        result['chi_ND'] = pd.DataFrame(CHI_ND)   # why dataframe?
        result['ZPF'] = PHI_zpf
        result['Pm_normed'] = PJ
        try:
            result['Pm_raw'] = self.PM[variation][self.PM[variation].columns[0]][modes]#TODO change the columns to junctions
        except:
             result['Pm_raw'] = self.PM[variation]
        _temp = self._get_participation_normalized(
            variation, _renorm_pj=self._renorm_pj, print_=print_result)
        result['_Pm_norm'] = _temp['Pm_norm'][modes]
        result['_Pm_cap_norm'] = _temp['Pm_cap_norm'][modes]

        # just propagate
        result['hfss_variables'] = self._hfss_variables[variation]
        result['Ljs'] = self.Ljs[variation]
        result['Cjs'] = self.Cjs[variation]
        try:
            result['Q_coupling'] = self.Qm_coupling[variation][self.Qm_coupling[variation].columns[junctions]][modes]#TODO change the columns to junctions
        except:
             result['Q_coupling'] = self.Qm_coupling[variation]
        
        try:
            result['Qs'] = self.Qs[variation][self.PM[variation].columns[junctions]][modes] #TODO change the columns to junctions
        except:
             result['Qs'] = self.Qs[variation][modes]
        result['fock_trunc'] = fock_trunc
        result['cos_trunc'] = cos_trunc

        self.results[variation] = result
        self.results.save()
        
        if print_result:
            self.print_variation(variation)
            self.print_result(result)
    
        self.n_modes = tmp_n_modes #TODO is this smart should consider defining the modes of intrest in the initilazaition of the quantum object
        self.modes[variation]=tmp_modes 
        return result
    def full_report_variations(self, var_list: list=None):
        """see full_variation_report"""
        if var_list is None: var_list =self.variations
        for variation in var_list: 
            self.full_variation_report(variation)
    
    def full_variation_report(self,variation):
        """
        prints the results and paramters of a specific variation

        Parameters
        ----------
        variation : int or str
            the variation to be printed .

        Returns
        -------
        None.

        """
        self.print_variation(variation)
        
        self.print_result(variation)
        
        
    def print_variation(self, variation):
        """
        Utility reporting function
        """
        if variation is int: variation = str(variation)

        if len(self.hfss_vars_diff_idx) > 0:
            print('\n*** Different parameters')
            display(self._hfss_variables[self.hfss_vars_diff_idx][variation])
            print('\n')

        print('*** P (participation matrix, not normlz.)')
        print(self.PM[variation])

        print('\n*** S (sign-bit matrix)')
        print(self.SM[variation])

    def print_result(self, result):
        """
        Utility reporting function
        """
        if type(result) is str or type(result) is int: result = self.results[str(result)]

        # TODO: actually make into dataframe with mode labela and junction labels
        pritm = lambda x, frmt="{:9.2g}": print_matrix(x, frmt=frmt)

        print('*** P (participation matrix, normalized.)')
        pritm(result['Pm_normed'])

        print('\n*** Chi matrix O1 PT (MHz)\n    Diag is anharmonicity, off diag is full cross-Kerr.')
        pritm(result['chi_O1'], "{:9.3g}")

        print('\n*** Chi matrix ND (MHz) ')
        pritm(result['chi_ND'], "{:9.3g}")

        print('\n*** Frequencies O1 PT (MHz)')
        print(result['f_1'])

        print('\n*** Frequencies ND (MHz)')
        print(result['f_ND'])

        print('\n*** Q_coupling')
        print(result['Q_coupling'])

    def plotting_dic_x(self, Var_dic, var_name):
        dic = {}

        if (len(Var_dic.keys())+1) == self.Num_hfss_vars_diff_idx:
            lv, lv_str = self.get_variation_of_multiple_variables_value(
                Var_dic)
            dic['label'] = lv_str
            dic['x_label'] = var_name
            dic['x'] = self.get_variable_value(var_name, lv=lv)
        else:
            raise ValueError('more than one hfss variablae changes each time')

        return lv, dic

    # Does not seem used. What is Var_dic and var_name going to?
    # def plotting_dic_data(self, Var_dic, var_name, data_name):
    #   lv, dic = self.plotting_dic_x()
    #    dic['y_label'] = data_name

    def plot_results(self, result, Y_label, variable, X_label, variations: list = None):
        # TODO?
        pass

    def plot_hamiltonian_results(self,
                                 swp_variable: str = 'variation',
                                 variations: list = None,
                                 fig=None,
                                 x_label: str = None):
        """Plot results versus variation

        Keyword Arguments:
            swp_variable {str} -- Variable against which we swept. If noen, then just
                                    take the variation index (default: {None})
            variations {list} -- [description] (default: {None})
            fig {[type]} -- [description] (default: {None})

        Returns:
            fig, axs
        """

        x_label = x_label or swp_variable

        # Create figure and axes
        if not fig:
            fig, axs = plt.subplots(2, 2, figsize=(10, 6))
        else:
            axs = fig.axs

        ############################################################################
        ### Axis: Frequencies
        f0 = self.results.get_frequencies_HFSS(
            variations=variations, vs=swp_variable).transpose()
        f1 = self.results.get_frequencies_O1(
            variations=variations, vs=swp_variable).transpose()
        f_ND = self.results.get_frequencies_ND(
            variations=variations, vs=swp_variable).transpose()
        # changed by Asaf from f0 as not all modes are always analyzed
        mode_idx = list(f1.columns)
        n_modes = len(mode_idx)

        ax = axs[0, 0]
        ax.set_title('Modal frequencies (MHz)')

        # TODO: shouldmove these kwargs to the config
        cmap = cmap_discrete(n_modes)
        kw = dict(ax=ax, color=cmap, legend=False, lw=0, ms=0)

        # Choose which freq should have the solid line drawn with it. ND if present, else f1
        if f_ND.empty:
            plt_me_line = f1
            markerf1 = 'o'
        else:
            plt_me_line = f_ND
            markerf1 = '.'

            # plot the ND as points if present
            f_ND.plot(**{**kw, **dict(marker='o', ms=4, zorder=30)})

        f0.plot(**{**kw, **dict(marker='x', ms=2, zorder=10)})
        f1.plot(**{**kw, **dict(marker=markerf1, ms=4, zorder=20)})
        plt_me_line.plot(**{**kw, **dict(lw=1, alpha=0.6, color='grey')})

        ############################################################################
        # Axis: Quality factors'
        Qs = self.get_quality_factors(swp_variable=swp_variable)
        Qs = Qs if variations is None else Qs[variations]
        Qs = Qs.transpose()

        ax = axs[1, 0]
        ax.set_title('Quality factors')
        Qs.plot(ax=ax, lw=0, marker=markerf1, ms=4,
                legend=True, zorder=20, color=cmap)
        Qs.plot(ax=ax, lw=1, alpha=0.2, color='grey', legend=False)
        ax.set_yscale('log')

        ############################################################################
        ### Axis: Alpha and chi

        axs[0][1].set_title('Anharmonicities (MHz)')
        axs[1][1].set_title('Cross-Kerr frequencies (MHz)')

        def plot_chi_alpha(chi, primary):
            """
            Intenral function to plot chi and then also to plot alpha
            """
            idx = pd.IndexSlice
            kw1 = dict(lw=0, ms=4,  marker='o' if primary else 'x')
            kw2 = dict(lw=1, alpha=0.2, color='grey', label='_nolegend_')
            # ------------------------
            # Plot anharmonicity
            ax = axs[0, 1]
            for i, mode in enumerate(mode_idx):  # mode index number, mode index
                alpha = chi.loc[idx[:, mode], mode].unstack(1)
                alpha.plot(ax=ax, label=mode, color=cmap[i], **kw1)
                if primary:
                    alpha.plot(ax=ax, **kw2)

            # ------------------------
            # Plot chi
            ax = axs[1, 1]
            for mode in mode_idx:  # mode index number, mode index
                # restart the color counter i; n= mode2
                for i, mode2 in enumerate(mode_idx):
                    if int(mode2) > int(mode):
                        chi_element = chi.loc[idx[:, mode], mode2].unstack(1)
                        chi_element.plot(
                            ax=ax, label=f"{mode},{mode2}", color=cmap[i], **kw1)

                        if primary:
                            chi_element.plot(ax=ax, **kw2)

        def do_legends():
            legend_translucent(axs[0][1],  leg_kw=dict(
                fontsize=7, title='Mode'))
            legend_translucent(axs[1][1],  leg_kw=dict(fontsize=7))

        chiO1 = self.get_chis(variations=variations,
                              swp_variable=swp_variable, numeric=False)
        chiND = self.get_chis(variations=variations,
                              swp_variable=swp_variable, numeric=True)

        use_ND = not np.any(
            [r['fock_trunc'] == None for k, r in self.results.items()])
        if use_ND:
            plot_chi_alpha(chiND, True)
            do_legends()
            plot_chi_alpha(chiO1, False)
        else:
            plot_chi_alpha(chiO1, True)
            do_legends()

        for ax1 in axs:
            for ax in ax1:
                ax.set_xlabel(x_label)

        # Wrap up
        fig.tight_layout()

        return fig, axs

    # Below are functions introduced in v0.8 and newer

    def report_results(self, swp_variable='variation', numeric=True):
        """
        Report in table form the results in a markdown friendly way in Jupyter notebook
        using the pandas interface.
        """

        with pd.option_context('display.precision', 2):
            display(Markdown(("#### Mode frequencies (MHz)")))
            display(Markdown(("###### Numerical diagonalization")))
            display(self.get_frequencies(
                swp_variable=swp_variable, numeric=numeric))

            display(Markdown(("#### Kerr Non-linear coefficient table (MHz)")))
            display(Markdown(("###### Numerical diagonalization")))
            display(self.get_chis(swp_variable=swp_variable, numeric=numeric))

    def get_chis(self, swp_variable='variation', numeric=True, variations: list = None,
                 m=None, n=None):
        """return as multiindex data table

        If you provide m and n as integers or mode labels, then the chi between these modes will
        be returned as a pandas Series.
        """
        label = 'chi_ND' if numeric else 'chi_O1'
        df = pd.concat(self.results.vs_variations(
            label, vs=swp_variable, variations=variations),
            names=[swp_variable])

        if m is None and n is None:
            return df
        else:
            s = df.loc[pd.IndexSlice[:, m], n].unstack(1)[m]
            return s

    def get_frequencies(self, swp_variable='variation', numeric=True, variations: list = None):
        """return as multiindex data table
            index: eigenmode label
            columns: variation label
        """
        label = 'f_1' if numeric else 'f_ND'
        return self.results.vs_variations(label, vs=swp_variable, to_dataframe=True, variations=variations)

    def get_quality_factors(self, swp_variable='variation', variations: list = None):
        """return as pd.Series
            index: eigenmode label
            columns: variation label
        """
        return self.results.vs_variations('Qs', vs=swp_variable, to_dataframe=True, variations=variations)

    def get_participations(self, swp_variable='variation',
                           variations: list = None,
                           inductive=True,
                           _normed=True):
        """

            inductive (bool): EPR forjunciton inductance when True, else for capactiors

        Returns:
        ----------------
        Returns a multindex dataframe:
            index 0: sweep variable
            index 1: mode number
            column: junction number

        Example use:
        ---------------
        Plot the participation ratio of all junctions for a given mode vs a sweep of Lj.

        .. code-block language:python

            df=epra.get_participations(swp_variable='Lj')
            df.loc[pd.IndexSlice[:,0],0].unstack(1).plot(marker='o')
        """

        if inductive:
            if _normed:
                getme = 'Pm_normed'
            else:
                getme = 'Pm_raw'
        else:
            if _normed:
                getme = 'Pm_cap'
            else:
                raise NotImplementedError(
                    'not inductive and not _normed not implemented')

        participations = self.results.vs_variations(getme, vs=swp_variable)

        p2 = OrderedDict()
        for key, val in participations.items():
            df = pd.DataFrame(val)
            df.index.name = 'mode'
            df.columns.name = 'junc_idx'
            p2[key] = df

        participations = pd.concat(p2, names=[swp_variable])

        return participations

    def _get_PM_as_DataFrame(self):
        """
        Pm = epra._get_PM_as_DataFrame()
        Pm.unstack(1).groupby(axis=1,level=1).plot()
        """
        Pm = pd.concat(self.PM)
        Pm.index.set_names(['variation', 'mode'], inplace=True)
        Pm.columns.set_names(['junction'], inplace=True)
        return Pm

    def get_ansys_energies(self, swp_var='variation'):
        """
        Return a multi-index dataframe of ansys energies vs swep_variable

        Args:
            swp_var (str) :
        """
        if swp_var == 'variation':
            energies = self.ansys_energies
        else:
            energies = self.get_vs_variable(swp_var, 'ansys_energies')

        df = pd.concat({k: pd.DataFrame(v).transpose()
                        for k, v in energies.items()})
        df.index.set_names([swp_var, 'mode'], inplace=True)

        return df

    def quick_plot_participation(self, mode, junction, swp_variable='variation', ax=None, kw=None):
        """Quick plot participation for one mode

            kw : extra plot arguments
        """
        df = self.get_participations(swp_variable=swp_variable)
        kw = kw or {}
        ax = ax or plt.gca()
        df.loc[pd.IndexSlice[:, mode], junction].unstack(
            1).plot(marker='o', ax=ax, **kw)
        ax.set_ylabel(f'p_({mode},{junction})')

    def quick_plot_frequencies(self, mode, swp_variable='variation', ax=None, kw=None, numeric=False):
        """Quick plot freq for one mode

            kw : extra plot arguments
        """
        kw = kw or {}
        ax = ax or plt.gca()

        s = self.get_frequencies(
            numeric=numeric, swp_variable=swp_variable).transpose()[mode]
        s.plot(marker='o', ax=ax, **kw)

        ax.set_ylabel(f'$\\omega_{mode}$ (MHz)')

    def quick_plot_chi_alpha(self, mode1, mode2, swp_variable='variation', ax=None, kw=None, numeric=False):
        """Quick plot chi between mode 1 and mode 2.

        If you select mode1=mode2, then you will plot the alpha

            kw : extra plot arguments
        """
        kw = kw or {}
        ax = ax or plt.gca()

        s = self.get_chis(swp_variable=swp_variable,
                          numeric=numeric).loc[pd.IndexSlice[:, mode1], mode2].unstack(1)
        s.plot(marker='o', ax=ax, **kw)

        if mode1 == mode2:
            ax.set_ylabel(f'$\\alpha$({mode1}) (MHz) [anharmonicity]')
        else:
            ax.set_ylabel(f'$\\chi$({mode1,mode2}) (MHz) [total split]')

    def quick_plot_mode(self, mode, junction, mode1=None, swp_variable='variation', numeric=False, sharex=True):
        r"""Create a quick report to see mode parameters for only a single mode and a
        cross-kerr coupling to another mode.
        Plots the participation and cross participation
        Plots the frequencie
        plots the anharmonicity

        The values are either for the numeric or the non-numeric results, set by `numeric`
        """

        fig, axs = plt.subplots(2, 2, figsize=(12*0.9, 7*0.9))
        self.quick_plot_frequencies(
            mode, swp_variable=swp_variable, numeric=numeric, ax=axs[0, 1])
        self.quick_plot_participation(
            mode, junction, swp_variable=swp_variable, ax=axs[0, 0])
        self.quick_plot_chi_alpha(mode, mode, numeric=numeric, swp_variable=swp_variable, ax=axs[1, 0],
                                  kw=dict(sharex=sharex))
        if mode1:
            self.quick_plot_chi_alpha(
                mode, mode1, numeric=numeric, swp_variable=swp_variable, ax=axs[1, 1])
            twinax = axs[0, 0].twinx()
            self.quick_plot_participation(mode1, junction, swp_variable=swp_variable, ax=twinax,
                                          kw=dict(alpha=0.7, color='maroon', sharex=sharex))

        for ax in np.ndarray.flatten(axs):
            ax.grid(alpha=0.2)

        axs[0, 1].set_title('Frequency (MHz)')
        axs[0, 0].set_title('Self- and cross-EPR')
        axs[1, 0].set_title('Anharmonicity')
        axs[1, 1].set_title('Cross-Kerr')

        fig.suptitle(f'Mode {mode}', y=1.025)
        fig.tight_layout()

    def quick_plot_convergence(self, ax = None):
        """
        Plot a report of the Ansys converngece vs pass number ona twin axis
        for the number of tets and the max delta frequency of the eignemode.
        """
        ax = ax or plt.gca()
        ax_t = ax.twinx()

        convergence_tets = self.get_convergences_tets_vs_pass()
        convergence_freq = self.get_convergences_max_delta_freq_vs_pass()

        convergence_freq.name = 'Δf'

        plot_convergence_max_df(ax, convergence_freq)
        plot_convergence_solved_elem(ax_t, convergence_tets)


def extract_dic(name=None, file_name=None):
    """#name is the name of the dictionry as saved in the npz file if it is None,
    the function will return a list of all dictionaries in the npz file
    file name is the name of the npz file"""
    with np.load(file_name, allow_pickle=True) as f:
        if name is None:
            return [f[i][()] for i in f.keys()]
        return [f[name][()]]



"""
Python (py) Energy-Participation-Ratio (EPR) package
pyEPR is an open source, BSD-licensed library providing high-efficiency,
easy-to-use analysis functions and automation for the design of quantum
chips based on superconducting quantum circuits, both distributed and lumped.
pyEPR interfaces the classical distributed microwave analysis with that of
quantum structures and Hamiltonians. It is chiefly based on the energy participation
ratio approach; however, it has since v0.4 extended to cover a broad range of
design approaches. pyEPR stradels the analysis from Maxwell’s to Schrodinger’s
equations, and converts the solutions of distributed microwve (typically eignmode
simulations) to a fully diagonalized spectrum of the energy levels, couplings,
and key parameters of a many-body quantum Hamiltonian.

Read the docs: https://pyepr-docs.readthedocs.io/en/latest/
Github page: https://github.com/zlatko-minev/pyEPR
"""

from pathlib import Path
from setuptools import setup, find_packages

here = Path(__file__).parent.absolute()

# Get the long description from the README file
with open(here / "README.md", encoding="utf-8") as f:
    long_description = f.read()

with open(here / "requirements.txt", encoding="utf-8") as f:
    requirements = f.read().splitlines()

doclines = __doc__.split('\n')

setup(name='pyEPR-quantum',
      version='0.8.4',
      description = doclines[0],
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Zlatko K. Minev',
      packages=find_packages(),
      author_email='zlatko.minev@aya.yale.edu',
      maintainer='Zlatko Minev, pyEPR team',
      license='BSD-3-Clause',
      url=r'https://github.com/zlatko-minev/pyEPR',
      classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License"],
      python_requires=">=3.5, <4",
      # install_requires=['numpy','pandas','pint','matplotlib','attrdict','sympy','IPython'],
      install_requires=requirements
      )



"""
Hamiltonian and Matrix Operations.
Hamiltonian operations heavily draw on qutip package.
This package must be installded for them to work.
"""
try:
    import qutip
    from qutip import Qobj  # basis, tensor,
except (ImportError, ModuleNotFoundError):
    Qobj=None
    pass

from ..toolbox.pythonic import fact


class MatrixOps(object):

    @staticmethod
    def cos(op_cos_arg: Qobj):
        """
        Make cosine opertor matrix from arguemnt  op_cos_arg

            op_cos_arg (qutip.Qobj) : argumetn of the cosine
        """

        return 0.5*((1j*op_cos_arg).expm() + (-1j*op_cos_arg).expm())

    @staticmethod
    def cos_approx(x, cos_trunc=5):
        """
        Create a Taylor series matrix approximation of the cosine, up to some order.
        """
        return sum((-1)**i * x**(2*i) / float(fact(2*i)) for i in range(2, cos_trunc + 1))

    @staticmethod
    def dot(ais, bis):
        """
        Dot product
        """
        return sum(ai*bi for ai, bi in zip(ais, bis))


class HamOps(object):

    @staticmethod
    def fock_state_on(d: dict, fock_trunc: int, N_modes: int):
        ''' d={mode number: # of photons} In the bare eigen basis
        '''
        # give me the value d[i]  or 0 if d[i] does not exist
        return qutip.tensor(*[qutip.basis(fock_trunc, d.get(i, 0))
                              for i in range(N_modes)])

    @staticmethod
    def closest_state_to(s: Qobj, energyMHz, evecs):
        """
        Returns the enery of the closest state to s
        """
        def distance(s2):
            return (s.dag() * s2[1]).norm()
        return max(zip(energyMHz, evecs), key=distance)

    @staticmethod
    def closest_state_to_idx(s: Qobj, evecs):
        """
        Returns the index
        """
        def distance(s2):
            return (s.dag() * s2[1]).norm()
        return max(zip(range(len(evecs)), evecs), key=distance)

    @staticmethod
    def identify_Fock_levels(fock_trunc: int, evecs,
                             N_modes=2,
                             Fock_max=4):
        """
        Return quantum numbers in terms of the undiagonalized eigenbasis.
        """
        #  to do: need to turn Fock_max into arb algo on each mdoe

        def fock_state_on(d):
            return HamOps.fock_state_on(d, fock_trunc, N_modes)

        def closest_state_to_idx(s):
            return HamOps.closest_state_to_idx(s, evecs)

        FOCKr = {}
        for d1 in range(Fock_max):
            for d2 in range(Fock_max):
                d = {0: d1, 1: d2}
                FOCKr[closest_state_to_idx(fock_state_on(d))[0]] = d
        return FOCKr



"""
Implementation of basic quantum operation in numpy,
to effortleslly remove the need in the `qutip` package.
"""

import numpy as np

def create(n: int):
    """Returns matrix representation of an n-dimensional creation operator"""
    diag = np.sqrt(np.arange(1,n))
    mat = np.zeros([n, n])
    np.fill_diagonal(mat[1:], diag)
    return mat

def destroy(n: int):
    """Returns matrix representation of an n-dimensional annhilation operator"""
    diag = np.sqrt(np.arange(1, n))
    mat = np.zeros([n, n])
    np.fill_diagonal(mat[:, 1:], diag)
    return mat

def number(n: int):
    """Returns matrix representation of an n-dimensional number operator"""
    mat = np.zeros([n, n])
    np.fill_diagonal(mat, np.arange(n))
    return mat

def basis(n: int, N: int): # Numpy does provide a method that does this but it's very slow
    """Returns the n-th, N-dimensional basis vector"""
    vec = np.zeros([N, 1])
    vec[n] = 1.0
    return vec



"""
Main interface module to use pyEPR.

Contains code to conenct to Ansys and to analyze HFSS files using the EPR method.

This module handles the micowave part of the analysis and conenction to

Further contains code to be able to do autogenerated reports,

Copyright Zlatko Minev, Zaki Leghtas, and the pyEPR team
2015, 2016, 2017, 2018, 2019, 2020
"""

from __future__ import print_function  # Python 2.7 and 3 compatibility

import sys
from pathlib import Path

import pandas as pd

from . import Dict, ansys, config, logger
from .toolbox.pythonic import get_instance_vars


diss_opt = ['dielectrics_bulk', 'dielectric_surfaces', 'resistive_surfaces', 'seams']

class ProjectInfo(object):
    """
    Primary class to store interface information between ``pyEPR`` and ``Ansys``.

    * **Ansys:** stores and provides easy access to the ansys interface classes :py:class:`pyEPR.ansys.HfssApp`,
      :py:class:`pyEPR.ansys.HfssDesktop`, :py:class:`pyEPR.ansys.HfssProject`, :py:class:`pyEPR.ansys.HfssDesign`,
      :py:class:`pyEPR.ansys.HfssSetup` (which, if present could nbe a subclass, such as a driven modal setup
      :py:class:`pyEPR.ansys.HfssDMSetup`, eigenmode :py:class:`pyEPR.ansys.HfssEMSetup`, or Q3D  :py:class:`pyEPR.ansys.AnsysQ3DSetup`),
      the 3D modeler to design geometry :py:class:`pyEPR.ansys.HfssModeler`.
    * **Junctions:** The class stores params about the design that the user puts will use, such as the names and
      properties of the junctions, such as whihc rectangle and line is associated with which junction.


    Note:

        **Junction parameters.**
        The junction parameters are stored in the ``self.junctions`` ordered dictionary

        A Josephson tunnel junction has to have its parameters specified here for the analysis.
        Each junction is given a name and is specified by a dictionary.
        It has the following properties:

        * ``Lj_variable`` (str):
                Name of HFSS variable that specifies junction inductance Lj defined
                on the boundary condition in HFSS.
                WARNING: DO NOT USE Global names that start with $.
        * ``rect`` (str):
                String of Ansys name of the rectangle on which the lumped boundary condition is defined.
        * ``line`` (str):
                Name of HFSS polyline which spans the length of the recntalge.
                Used to define the voltage across the junction.
                Used to define the current orientation for each junction.
                Used to define sign of ZPF.
        * ``length`` (str):
                Length in HFSS of the junction rectangle and line (specified in meters).
                To create, you can use :code:`epr.parse_units('100um')`.
        * ``Cj_variable`` (str, optional) [experimental]:
                Name of HFSS variable that specifies junction inductance Cj defined
                on the boundary condition in HFSS. DO NOT USE Global names that start with ``$``.

    Warning:

        To define junctions, do **NOT** use global names!
        I.e., do not use names in ansys that start with ``$``.


    Note:

        **Junction parameters example .** To define junction parameters, see the following example

        .. code-block:: python
            :linenos:

            # Create project infor class
            pinfo = ProjectInfo()

            # Now, let us add a junction called `j1`, with the following properties
            pinfo.junctions['j1'] = {
                        'Lj_variable' : 'Lj_1', # name of Lj variable in Ansys
                        'rect'        : 'jj_rect_1',
                        'line'        : 'jj_line_1',
                        #'Cj'          : 'Cj_1' # name of Cj variable in Ansys - optional
                        }

        To extend to define 5 junctions in bulk, we could use the following script

        .. code-block:: python
            :linenos:

            n_junctions = 5
            for i in range(1, n_junctions + 1):
                pinfo.junctions[f'j{i}'] = {'Lj_variable' : f'Lj_{i}',
                                            'rect'        : f'jj_rect_{i}',
                                            'line'        : f'jj_line_{i}'}


    .. _Google Python Style Guide:
        http://google.github.io/styleguide/pyguide.html

    """

    class _Dissipative:
        """
        Deprecating the _Dissipative class and turning it into a dictionary.
        This is used to message people on the deprecation so they could change their scripts.
        """
        def __init__(self):
            self['pinfo'] = None
            for opt in diss_opt:
                self[opt] = None

        def __setitem__(self, key, value):
            # --- check valid inputs ---
            if not (key in diss_opt or key == 'pinfo'):
                raise ValueError(f"No such parameter {key}")
            if key != 'pinfo' and (not isinstance(value, list) or \
                    not all(isinstance(x, str) for x in value)) and (value != None):
                raise ValueError(f'dissipative[\'{key}\'] must be a list of strings ' \
                    'containing names of models in the project!')
            if key != 'pinfo' and hasattr(self['pinfo'], 'design'):
                for x in value:
                    if x not in self['pinfo'].get_all_object_names():
                        raise ValueError(
                            f'\'{x}\' is not an object in the HFSS project')
            super().__setattr__(key, value)

        def __getitem__(self, attr):
            if not (attr in diss_opt or attr == 'pinfo'):
                raise AttributeError(f'dissipitive has no attribute "{attr}". '\
                    f'The possible attributes are:\n {str(diss_opt)}')
            return super().__getattribute__(attr)

        def __setattr__(self, attr, value):
            logger.warning(
                f"DEPRECATED!! use pinfo.dissipative['{attr}'] = {value} instead!")
            self[attr] = value

        def __getattr__(self, attr):
            raise AttributeError(f'dissipitive has no attribute "{attr}". '\
                f'The possible attributes are:\n {str(diss_opt)}')

        def __getattribute__(self, attr):
            if attr in diss_opt:
                logger.warning(
                    f"DEPRECATED!! use pinfo.dissipative['{attr}'] instead!")
            return super().__getattribute__(attr)

        def __repr__(self):
            return str(self.data())

        def data(self):
            """Return dissipatvie as dictionary"""
            return {str(opt): self[opt] for opt in diss_opt}

    def __init__(self, project_path: str = None, project_name: str = None, design_name: str = None,
                 setup_name: str = None, do_connect: bool = True):
        """
        Keyword Arguments:

            project_path (str) : Directory path to the hfss project file.
                Should be the directory, not the file.
                Defaults to ``None``; i.e., assumes the project is open, and thus gets the project based
                on `project_name`.
            project_name (str) : Name of the project within the project_path.
                Defaults to ``None``, which will get the current active one.
            design_name  (str) : Name of the design within the project.
                Defaults to ``None``, which will get the current active one.
            setup_name  (str) :  Name of the setup within the design.
                Defaults to ``None``, which will get the current active one.

            do_connect (bool) [additional]: Do create connection to Ansys or not? Defaults to ``True``.

        """

        # Path: format path correctly to system convention
        self.project_path = str(Path(project_path)) \
            if not (project_path is None) else None
        self.project_name = project_name
        self.design_name = design_name
        self.setup_name = setup_name

        # HFSS desgin: describe junction parameters
        # TODO: introduce modal labels
        self.junctions = Dict()  # See above for help
        self.ports = Dict()

        # Dissipative HFSS volumes and surfaces
        self.dissipative = self._Dissipative()
        self.options = config.ansys

        # Conected to HFSS variable
        self.app = None
        self.desktop = None
        self.project = None
        self.design = None
        self.setup = None

        if do_connect:
            self.connect()
            self.dissipative['pinfo'] = self

    _Forbidden = ['app', 'design', 'desktop', 'project',
                  'dissipative', 'setup', '_Forbidden', 'junctions']

    def save(self):
        '''
        Return all the data in a dectionary form that can be used to be saved
        '''
        return dict(
            pinfo=pd.Series(get_instance_vars(self, self._Forbidden)),
            dissip=pd.Series(self.dissipative.data()),
            options=pd.Series(get_instance_vars(self.options)),
            junctions=pd.DataFrame(self.junctions),
            ports=pd.DataFrame(self.ports),
        )

    def connect(self):
        """
        Do establihs connection to Ansys desktop.
        """
        logger.info('Connecting to Ansys Desktop API...')

        self.app, self.desktop, self.project = ansys.load_ansys_project(
            self.project_name, self.project_path)
        self.project_name = self.project.name
        self.project_path = self.project.get_path()

        # Design
        if self.design_name is None:
            self.design = self.project.get_active_design()
            self.design_name = self.design.name
            logger.info(f'\tOpened active design\n\
\tDesign:    {self.design_name} [Solution type: {self.design.solution_type}]')
        else:

            try:
                self.design = self.project.get_design(self.design_name)
                logger.info(f'\tOpened active design\n\
\tDesign:    {self.design_name} [Solution type: {self.design.solution_type}]')

            except Exception as e:
                _traceback = sys.exc_info()[2]
                logger.error(f"Original error \N{loudly crying face}: {e}\n")
                raise(Exception(' Did you provide the correct design name?\
                    Failed to pull up design. \N{loudly crying face}').with_traceback(_traceback))

        # Setup
        try:
            setup_names = self.design.get_setup_names()

            if len(setup_names) == 0:
                logger.warning('\tNo design setup detected.')
                if self.design.solution_type == 'Eigenmode':
                    logger.warning('\tCreating eigenmode default setup one.')
                    setup = self.design.create_em_setup()
                    self.setup_name = setup.name
                elif self.design.solution_type == 'DrivenModal':
                    setup = self.design.create_dm_setup()  # adding a driven modal design
                    self.setup_name = setup.name
            else:
                self.setup_name = setup_names[0]

            # get the actual setup if there is one
            self.get_setup(self.setup_name)

        except Exception as e:

            _traceback = sys.exc_info()[2]
            logger.error(f"Original error \N{loudly crying face}: {e}\n")
            raise Exception(' Did you provide the correct setup name?\
                        Failed to pull up setup. \N{loudly crying face}').with_traceback(_traceback)

        # Finalize
        self.project_name = self.project.name
        self.design_name = self.design.name

        logger.info(
            '\tConnection to Ansys established successfully. \N{grinning face} \n')

        return self

    def get_setup(self, name: str):
        """
        Connects to a specific setup for the design.
        Sets  self.setup and self.setup_name.

        Args:
            name (str): Name of the setup.
            If the setup does not exist, then throws a loggger error.
            Defaults to ``None``, in which case returns None

        """
        if name is None:
            return None
        else:
            self.setup = self.design.get_setup(name=self.setup_name)
            if self.setup is None:
                logger.error(f"Could not retrieve setup: {self.setup_name}\n \
                               Did you give the right name? Does it exist?")

            self.setup_name = self.setup.name
            logger.info(
                f'\tOpened setup `{self.setup_name}`  ({type(self.setup)})')
            return self.setup

    def check_connected(self):
        """
        Checks if fully connected including setup.
        """
        return\
            (self.setup is not None) and\
            (self.design is not None) and\
            (self.project is not None) and\
            (self.desktop is not None) and\
            (self.app is not None)

    def disconnect(self):
        '''
        Disconnect from existing HFSS design.
        '''
        assert self.check_connected() is True,\
            "It does not appear that you have connected to HFSS yet.\
            Use the connect()  method. \N{nauseated face}"
        self.project.release()
        self.desktop.release()
        self.app.release()
        ansys.release()

    # UTILITY FUNCTIONS

    def get_dm(self):
        '''
        Utility shortcut function to get the design and modeler.

        .. code-block:: python

            oDesign, oModeler = pinfo.get_dm()

        '''
        return self.design, self.design.modeler

    def get_all_variables_names(self):
        """Returns array of all project and local design names."""
        return self.project.get_variable_names() + self.design.get_variable_names()

    def get_all_object_names(self):
        """Returns array of strings"""
        o_objects = []
        for s in ["Non Model", "Solids", "Unclassified", "Sheets", "Lines"]:
            o_objects += self.design.modeler.get_objects_in_group(s)
        return o_objects

    def validate_junction_info(self):
        """Validate that the user has put in the junction info correctly.
        Do not also forget to check the length of the rectangles/line of
        the junction if you change it.
        """

        all_variables_names = self.get_all_variables_names()
        all_object_names = self.get_all_object_names()

        for jjnm, jj in self.junctions.items():

            assert jj['Lj_variable'] in all_variables_names,\
                """pyEPR ProjectInfo user error found \N{face with medical mask}:
                Seems like for junction `%s` you specified a design or project
                variable for `Lj_variable` that does not exist in HFSS by the name:
                 `%s` """ % (jjnm, jj['Lj_variable'])

            for name in ['rect', 'line']:

                assert jj[name] in all_object_names, \
                    """pyEPR ProjectInfo user error found \N{face with medical mask}:
                    Seems like for junction `%s` you specified a %s that does not exist
                    in HFSS by the name: `%s` """ % (jjnm, name, jj[name])
  
    def __del__(self):
        logger.info('Disconnected from Ansys HFSS')
        self.disconnect()



"""
Main distributed analysis module to use pyEPR.

Contains code to conenct to Ansys and to analyze HFSS files using the EPR method.

This module handles the micowave part of the analysis and conenction to

Further contains code to be able to do autogenerated reports,

Copyright Zlatko Minev, Zaki Leghtas, and the pyEPR team
2015, 2016, 2017, 2018, 2019, 2020
"""
# pylint: disable=invalid-name
# todo remove this pylint hack later

from __future__ import print_function  # Python 2.7 and 3 compatibility

from typing import List

import pickle
import sys
import time
from collections import OrderedDict
from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from . import Dict, config, logger
from .ansys import CalcObject, ConstantVecCalcObject, set_property, ureg
from .calcs.constants import epsilon_0
from .project_info import ProjectInfo
from .reports import (plot_convergence_f_vspass, plot_convergence_max_df,
                      plot_convergence_maxdf_vs_sol,
                      plot_convergence_solved_elem)
from .toolbox.pythonic import print_NoNewLine


# class AnsysAnalysisBase():
#
#    def __init__():
#        """
#        Instantiate in super after call.
#        """


class DistributedAnalysis(object):
    """
    DISTRIBUTED ANALYSIS of layout and microwave results.

    Main compuation class & interface with HFSS.

    This class defines a DistributedAnalysis object which calculates and saves
    Hamiltonian parameters from an HFSS simulation.

    Further, it allows one to calcualte dissipation, etc.
    """

    def __init__(self, *args, **kwargs):
        '''
        Pass in the arguments for ProjectInfo. See help for `?ProjectInfo`.


        Parameters:
        -------------------
            project_info    : ProjectInfo
                Suplpy the project info or the parameters to create pinfo

        Use notes:
        -------------------
            * If you change the setup or number of eignemodes in HFSS, etc.
              call `update_ansys_info()`


        Example use:
        -------------------

        See the tutorials in the repository.

        .. code-block:: python
            :linenos:

            import pyEPR as epr
            pinfo = epr.ProjectInfo(project_path = path_to_project,
                                    project_name = 'pyEPR_tutorial1',
                                    design_name  = '1. single_transmon')
            eprd = epr.DistributedAnalysis(pinfo)

        To now quickly see the result of a sweep of a variable in ansys, you can use:

        .. code-block:: python
            :linenos:

            swp_var = 'Lj'
            display(eprd.get_ansys_variables())
            fs = eprd.quick_plot_frequencies(swp_var)
            display(fs)

        To perform distributed analysis

        .. code-block:: python
            :linenos:

            eprd.do_EPR_analysis(append_analysis=True);

        Key internal paramters:
        -------------------
            n_modes (int) : Number of eignemodes; e.g., 2
            variations (List[str]) :  A list of string identifier of **solved** variation
                for the selected setup. Example: '['0', '1']
            _list_variations :  An array of strings corresponding to **solved** variations.
                List of identifier strings for the SOLVED ansys variation for the selected setup.
                These do not include unsolved variables added after the solution!

                .. code-block:: python

                             ("Height='0.06mm' Lj='13.5nH'",   "Height='0.06mm' Lj='15.3nH'")

        '''

        # Get the project info
        project_info = None
        if (len(args) == 1) and (args[0].__class__.__name__ == 'ProjectInfo'):
            # isinstance(args[0], ProjectInfo): # fails on module repload with changes
            project_info = args[0]
        else:
            assert len(args) == 0, '''Since you did not pass a ProjectInfo object
                as a arguemnt, we now assuem you are trying to create a project
                info object here by apassing its arguments. See ProjectInfo.
                It does not take any arguments, only kwargs. \N{face with medical mask}'''
            project_info = ProjectInfo(*args, **kwargs)

        # Input
        self.pinfo = project_info  # : project_info: a reference to a Project_Info class
        if self.pinfo.check_connected() is False:
            self.pinfo.connect()

        # hfss connect module
        self.fields = None
        self.solutions = None
        if self.setup:
            self.fields = self.setup.get_fields()
            self.solutions = self.setup.get_solutions()

        # Stores resutls from sims
        self.results = Dict()  # of variations. Saved results
        # TODO: turn into base class shared with analysis!

        # Modes and variations - the following get updated in update_variation_information
        self.n_modes = int(1)  # : Number of eigenmodes
        self.modes = None
        #: List of variation indecies, which are strings of ints, such as ['0', '1']
        self.variations = []
        self.variations_analyzed = []  # : List of analyzed variations. List of indecies

        # String identifier of variables, such as  "Cj='2fF' Lj='12.5nH'"
        self._nominal_variation = ''
        self._list_variations = ("",)  # tuple set of variables
        # container for eBBQ list of varibles; basically the same as _list_variations
        self._hfss_variables = Dict()

        self._previously_analyzed = set()  # previously analyzed variations

        self.update_ansys_info()

        print('Design \"%s\" info:' % self.design.name)
        print('\t%-15s %d\n\t%-15s %d' % ('# eigenmodes', self.n_modes,
                                          '# variations', self.n_variations))

        # Setup data saving
        self.data_dir = None
        self.file_name = None
        self.setup_data()

    @property
    def setup(self):
        """Ansys setup class handle. Could be None."""
        return self.pinfo.setup

    @property
    def design(self):
        """Ansys design class handle"""
        return self.pinfo.design

    @property
    def project(self):
        """Ansys project class handle"""
        return self.pinfo.project

    # @property
    # def desktop(self):
    #    """Ansys desktop class handle"""
    #    return self.pinfo.desktop

    # @property
    # def app(self):
    #    """Ansys App class handle"""
    #    return self.pinfo.app

    # @property
    # def junctions(self):
    #    """Project info junctions"""
    #    return self.pinfo.junctions

    # @property
    # def ports(self):
    #    return self.pinfo.ports

    @property
    def options(self):
        """ Project info options"""
        return self.pinfo.options

    def setup_data(self):
        '''
        Set up folder paths for saving data to.

        Sets the save filename with the current time.

        Saves to Path(config.root_dir) / self.project.name / self.design.name
        '''

        if len(self.design.name) > 50:
            logger.error('WARNING!   DESIGN FILENAME MAY BE TOO LONG! ')

        self.data_dir = Path(config.root_dir) / \
            self.project.name / self.design.name
        self.data_filename = self.data_dir / (time.strftime(config.save_format,
                                                            time.localtime()) + '.npz')

        if not self.data_dir.is_dir():
            self.data_dir.mkdir(parents=True, exist_ok=True)

    def calc_p_junction_single(self, mode, variation, U_E=None, U_H=None):
        '''
        This function is used in the case of a single junction only.
        For multiple junctions, see `calc_p_junction`.

        Assumes no lumped capacitive elements.
        '''
        if U_E is None:
            U_E = self.calc_energy_electric(variation)
        if U_H is None:
            U_H = self.calc_energy_magnetic(variation)

        pj = OrderedDict()
        pj_val = (U_E-U_H)/U_E
        pj['pj_'+str(mode)] = np.abs(pj_val)
        print('    p_j_' + str(mode) + ' = ' + str(pj_val))
        return pj

    # TODO: replace this method with the one below, here because osme funcs use it still
    def get_freqs_bare(self, variation: str):
        """
        Warning:
            Outdated. Do not use. To be depreicated

        Args:
            variation (str): A string identifier of the variation,
                such as '0', '1', ...

        Returns:
            [type] -- [description]
        """
        # str(self._get_lv(variation))
        freqs_bare_vals = []
        freqs_bare_dict = OrderedDict()
        freqs, kappa_over_2pis = self.solutions.eigenmodes(
            self.get_variation_string(variation))
        for m in range(self.n_modes):
            freqs_bare_dict['freq_bare_'+str(m)] = 1e9*freqs[m]
            freqs_bare_vals.append(1e9*freqs[m])
            if kappa_over_2pis is not None:
                freqs_bare_dict['Q_'+str(m)] = freqs[m]/kappa_over_2pis[m]
            else:
                freqs_bare_dict['Q_'+str(m)] = 0
        #self.freqs_bare = freqs_bare_dict
        #self.freqs_bare_vals = freqs_bare_vals
        return freqs_bare_dict, freqs_bare_vals

    def get_freqs_bare_pd(self, variation: str, frame=True):
        """Return the freq and Qs of the solved modes for a variation.
        I.e., the Ansys solved frequencies.



        Args:
            variation (str): A string identifier of the variation,
                such as '0', '1', ...
            frame {bool} -- if True returns dataframe, else tuple of series.

        Returns:
           If frame = True, then a multi-index Dataframe that looks something like this

            .. code-block:: python

                                Freq. (GHz)  Quality Factor
                variation mode
                0         0        5.436892             1020
                        1        7.030932             50200
                1         0        5.490328             2010
                        1        7.032116             104500

           If frame = False, then a tuple of  two Series, such as
           (Fs, Qs) -- Tuple of pandas.Series objects; the row index is the mode number
        """
        variation_str = self.get_variation_string(variation)

        freqs, kappa_over_2pis = self.solutions.eigenmodes(variation_str)
        if kappa_over_2pis is None:
            kappa_over_2pis = np.zeros(len(freqs))

        freqs = pd.Series(freqs, index=range(len(freqs)))  # GHz
        Qs = freqs / pd.Series(kappa_over_2pis, index=range(len(freqs)))

        if frame:
            df = pd.DataFrame({'Freq. (GHz)': freqs, 'Quality Factor': Qs})
            df.index.name = 'mode'
            return df
        else:
            return freqs, Qs

    def get_ansys_frequencies_all(self, vs='variation'):
        """
        Return all ansys frequencies and quality factors vs a variation

        Returns a multi-index pandas DataFrame
        """
        df = dict()
        variable = None if vs == 'variation' else self.get_variable_vs_variations(
            vs)
        for variation in self.variations:  # just for the first 2
            if vs == 'variation':
                label = variation
            else:
                label = variable[variation]
            df[label] = self.get_freqs_bare_pd(variation=variation)
        # TODO: maybe sort column and index? # todo: maybe generalize
        return pd.concat(df, names=[vs])

    def _get_lv(self, variation=None):
        '''
        List of variation variables in a format that is used when feeding back to ansys.

        Args:
            variation (str): A string identifier of the variation,
                such as '0', '1', ...

        Returns:
            list of var names and var values.
            Such as

            .. code-block:: python

                ['Lj1:=','13nH', 'QubitGap:=','100um']
        '''

        if variation is None:
            lv = self._nominal_variation  # "Cj='2fF' Lj='12.5nH'"
            lv = self._parse_listvariations(lv)
        else:
            lv = self._list_variations[ureg(variation)]
            lv = self._parse_listvariations(lv)
        return lv

    # Functions that deal with variations exclusively

    @property
    def n_variations(self):
        """ Number of **solved** variaitons, corresponding to the
        selected Setup. """
        return len(self._list_variations)

    def set_variation(self, variation: str):
        """
        Set the ansys design to a solved variation.
        This will change all local variables!

        Warning: not tested with global variables.
        """
        variation_string = self.get_variation_string(variation)
        self.design.set_variables(variation_string)

    def get_variations(self):
        """
        An array of strings corresponding to **solved** variations corresponding to the
        selected Setup.

        Returns:
            Returns a list of strings that give the variation labels for HFSS.
            .. code-block:: python

                OrderedDict([
                    ('0', "Cj='2fF' Lj='12nH'"),
                    ('1', "Cj='2fF' Lj='12.5nH'"),
                    ('2', "Cj='2fF' Lj='13nH'"),
                    ('3', "Cj='2fF' Lj='13.5nH'"),
                    ('4', "Cj='2fF' Lj='14nH'")])
        """
        return OrderedDict(zip(self.variations, self._list_variations))

    def get_variation_string(self, variation=None):
        """
        **Solved** variation string identifier.

        Args:
            variation (str): A string identifier of the variation,
                such as '0', '1', ...

        Returns:
            Return the list variation string of parameters in ansys used to identify the variation.

            .. code-block:: python

                "$test='0.25mm' Cj='2fF' Lj='12.5nH'"
        """
        if variation is None:
            return self._nominal_variation

        return self._list_variations[ureg(variation)]

    def _parse_listvariations(self, lv):
        """
        Turns
            "Cj='2fF' Lj='13.5nH'"
        into
            ['Cj:=', '2fF', 'Lj:=', '13.5nH']
        """
        lv = str(lv)
        lv = lv.replace("=", ":=,")
        lv = lv.replace(' ', ',')
        lv = lv.replace("'", "")
        lv = lv.split(",")
        return lv

    def get_nominal_variation_index(self):
        """
        Returns:
            A string identifies, such as '0' or '1', that labels the
            nominal variation index number.

        This may not be in the solved list!s
        """
        try:
            return str(self._list_variations.index(self._nominal_variation))
        except Exception:
            print('WARNING: Unsure of the index, returning 0')
            return '0'

    def get_ansys_variations(self):
        """
        Will update ansys inofrmation and result the list of variations.

        Returns:
            For example:

            .. code-block:: python

                ("Cj='2fF' Lj='12nH'",
                "Cj='2fF' Lj='12.5nH'",
                "Cj='2fF' Lj='13nH'",
                "Cj='2fF' Lj='13.5nH'",
                "Cj='2fF' Lj='14nH'")
        """
        self.update_ansys_info()
        return self._list_variations

    def update_ansys_info(self):
        ''''
        Updates all information about the Ansys solved variations and variables.

        .. code-block:: python
            :linenos:

            n_modes, _list_variations, nominal_variation, n_variations

        '''

        # from oDesign
        self._nominal_variation = self.design.get_nominal_variation()

        if self.setup:
            # from oSetup -- only for the solved variations!
            self._list_variations = self.solutions.list_variations()

            self.variations = [str(i) for i in range(
                self.n_variations)]  # TODO: change to integer?

            # eigenmodes
            if self.design.solution_type == 'Eigenmode':
                self.n_modes = int(self.setup.n_modes)
            else:
                self.n_modes = 0

        self._update_ansys_variables()

    def _update_ansys_variables(self, variations=None):
        """
        Updates the list of ansys hfss variables for the set of sweeps.
        """
        variations = variations or self.variations
        for variation in variations:
            self._hfss_variables[variation] = pd.Series(
                self.get_variables(variation=variation))
        return self._hfss_variables

    def get_ansys_variables(self):
        """
        Get ansys variables for all variaitons

        Returns:
            Return a dataframe of variables as index and columns as the variations
        """
        vs = 'variation'
        df = pd.DataFrame(self._hfss_variables, columns=self.variations)
        df.columns.name = vs
        df.index = [x[1:] if x.startswith('_') else x for x in df.index]
        #df.index.name = 'variable'
        return df

    def get_variables(self, variation=None):
        """
        Get ansys variables.

        Args:
            variation (str): A string identifier of the variation,
                such as '0', '1', ...
        """
        lv = self._get_lv(variation)
        variables = OrderedDict()
        for ii in range(int(len(lv)/2)):
            variables['_'+lv[2*ii][:-2]] = lv[2*ii+1]
        #self.variables = variables
        return variables

    def get_variable_vs_variations(self, variable: str, convert: bool = True):
        """
        Get ansys variables

        Return HFSS variable from self.get_ansys_variables() as a
        pandas series vs variations.

        Args:
            convert (bool) : Convert to a numeric quantity if possible using the
                        ureg
        """
        # TODO: These should be common function to the analysis and here!
        # BOth should be subclasses of a base class
        s = self.get_ansys_variables().loc[variable, :]  # : pd.Series
        if convert:
            s = s.apply(lambda x: ureg.Quantity(x).magnitude)
        return s

    def calc_energy_electric(self,
                             variation: str = None,
                             obj: str = 'AllObjects',
                             volume: str = 'Deprecated',
                             smooth: bool = False,
                             obj_dims: int = 3):
        r'''
        Calculates two times the peak electric energy, or 4 times the RMS,
        :math:`4*\mathcal{E}_{\mathrm{elec}}`
        (since we do not divide by 2 and use the peak phasors).

        .. math::
            \mathcal{E}_{\mathrm{elec}}=\frac{1}{4}\mathrm{Re}\int_{V}\mathrm{d}v\vec{E}_{\text{max}}^{*}\overleftrightarrow{\epsilon}\vec{E}_{\text{max}}

        Args:
            variation (str): A string identifier of the variation,
                such as '0', '1', ...
            obj (string | 'AllObjects'): Name of the object to integrate over
            smooth (bool | False) : Smooth the electric field or not when performing calculation
            obj_dims (int | 3) : 1 - line, 2 - surface, 3 - volume. Default volume

        Example:
            Example use to calcualte the energy participation ratio (EPR) of a substrate

            .. code-block:: python
                :linenos:

                ℰ_total  = epr_hfss.calc_energy_electric(obj='AllObjects')
                ℰ_substr = epr_hfss.calc_energy_electric(obj='Box1')
                print(f'Energy in substrate = {100*ℰ_substr/ℰ_total:.1f}%')

        '''
        if volume != 'Deprecated':
            logger.warning('The use of the "volume" argument is deprecated... use "obj" instead')
            obj = volume

        calcobject = CalcObject([], self.setup)

        vecE = calcobject.getQty("E")
        if smooth:
            vecE = vecE.smooth()
        A = vecE.times_eps()
        B = vecE.conj()
        A = A.dot(B)
        A = A.real()

        if obj_dims == 1:
            A = A.integrate_line(name=obj)
        elif obj_dims == 2:
            A = A.integrate_surf(name=obj)
        elif obj_dims == 3:
            A = A.integrate_vol(name=obj)
        else:
            logger.warning('Invalid object dimensions %s, using default of 3 (volume)' % obj_dims)
            A = A.integrate_vol(name=obj)

        lv = self._get_lv(variation)
        return A.evaluate(lv=lv)

    def calc_energy_magnetic(self,
                             variation: str = None,
                             obj: str = 'AllObjects',
                             volume: str = 'Deprecated',
                             smooth: bool = False,
                             obj_dims: int = 3):
        '''
        See calc_energy_electric.

        Args:
            variation (str): A string identifier of the variation,
                such as '0', '1', ...
            volume (string | 'AllObjects'): Name of the volume to integrate over
            smooth (bool | False) : Smooth the electric field or not when performing calculation
            obj_dims (int | 3) : 1 - line, 2 - surface, 3 - volume. Default volume
        '''
        if volume != 'Deprecated':
            logger.warning('The use of the "volume" argument is deprecated... use "obj" instead')
            obj = volume

        calcobject = CalcObject([], self.setup)

        vecH = calcobject.getQty("H")
        if smooth:
            vecH = vecH.smooth()
        A = vecH.times_mu()
        B = vecH.conj()
        A = A.dot(B)
        A = A.real()

        if obj_dims == 1:
            A = A.integrate_line(name=obj)
        elif obj_dims == 2:
            A = A.integrate_surf(name=obj)
        elif obj_dims == 3:
            A = A.integrate_vol(name=obj)
        else:
            logger.warn(f'Invalid object dimensions {obj_dims}, using default of 3 (volume)')
            A = A.integrate_vol(name=obj)

        lv = self._get_lv(variation)
        return A.evaluate(lv=lv)

    def calc_p_electric_volume(self,
                               name_dielectric3D,
                               relative_to='AllObjects',
                               E_total=None
                               ):
        r'''
        Calculate the dielectric energy-participatio ratio
        of a 3D object (one that has volume) relative to the dielectric energy of
        a list of object objects.

        This is as a function relative to another object or all objects.

        When all objects are specified, this does not include any energy
        that might be stored in any lumped elements or lumped capacitors.

        Returns:
        ---------
            ℰ_object/ℰ_total, (ℰ_object, _total)
        '''

        if E_total is None:
            logger.debug('Calculating ℰ_total')
            ℰ_total = self.calc_energy_electric(obj=relative_to)
        else:
            ℰ_total = E_total

        logger.debug('Calculating ℰ_object')
        ℰ_object = self.calc_energy_electric(obj=name_dielectric3D)

        return ℰ_object/ℰ_total, (ℰ_object, ℰ_total)

    def calc_current(self, fields, line: str):
        '''
        Function to calculate Current based on line. Not in use.

        Args:
            line (str) : integration line between plates - name
        '''
        self.design.Clear_Field_Clac_Stack()
        comp = fields.Vector_H
        exp = comp.integrate_line_tangent(line)
        I = exp.evaluate(phase=90)
        self.design.Clear_Field_Clac_Stack()
        return I

    def calc_avg_current_J_surf_mag(self, variation: str, junc_rect: str, junc_line):
        ''' Peak current I_max for mdoe J in junction J
            The avg. is over the surface of the junction. I.e., spatial.
        Args:
            variation (str): A string identifier of the variation,
                such as '0', '1', ...
            junc_rect (str) : name of rectangle to integrate over
            junc_line (str) : name of junction line to integrate over
        Returns:
            Value of peak current
        '''
        lv = self._get_lv(variation)

        jl, uj = self.get_junc_len_dir(variation, junc_line)

        uj = ConstantVecCalcObject(uj, self.setup)
        calc = CalcObject([], self.setup)
        #calc = calc.getQty("Jsurf").mag().integrate_surf(name = junc_rect)
        calc = (((calc.getQty("Jsurf")).dot(uj)).imag()
                ).integrate_surf(name=junc_rect)
        I = calc.evaluate(lv=lv) / jl  # phase = 90
        # self.design.Clear_Field_Clac_Stack()
        return I

    def calc_current_using_line_voltage(self, variation: str, junc_line_name: str,
                                        junc_L_Henries: float, Cj_Farads: float = None):
        '''
        Peak current I_max for prespecified mode calculating line voltage across junction.

        Make sure that oyu have set the correct variaitonin hFSS before running this

        Parameters:
        ------------------------------------------------
            variation: variation number
            junc_line_name: name of the HFSS line spanning the junction
            junc_L_Henries: junction inductance in henries
            Cj_Farads : junction cap in Farads
            TODO: Smooth?
        '''
        lv = self._get_lv(variation)
        v_calc_real = CalcObject([], self.setup).getQty(
            "E").real().integrate_line_tangent(name=junc_line_name)
        v_calc_imag = CalcObject([], self.setup).getQty(
            "E").imag().integrate_line_tangent(name=junc_line_name)
        V = np.sqrt(v_calc_real.evaluate(lv=lv)**2 +
                    v_calc_imag.evaluate(lv=lv)**2)

        # Get frequency
        freq = CalcObject(
            [('EnterOutputVar', ('Freq', "Complex"))], self.setup).real().evaluate()
        omega = 2*np.pi*freq  # in SI radian Hz units

        Z = omega*junc_L_Henries
        if abs(float(Cj_Farads)) > 1E-29:  # zero
            #print('Non-zero Cj used in calc_current_using_line_voltage')
            #Z += 1./(omega*Cj_Farads)
            print(
                '\t\t'f'Energy fraction (Lj over Lj&Cj)= {100./(1.+omega**2 *Cj_Farads*junc_L_Henries):.2f}%')
            # f'Z_L= {omega*junc_L_Henries:.1f} Ohms Z_C= {1./(omega*Cj_Farads):.1f} Ohms')

        I_peak = V/Z  # I=V/(wL)s

        return I_peak, V, freq

    def calc_line_current(self, variation, junc_line_name):
        lv = self._get_lv(variation)
        calc = CalcObject([], self.setup)
        calc = calc.getQty("H").imag().integrate_line_tangent(
            name=junc_line_name)
        # self.design.Clear_Field_Clac_Stack()
        return calc.evaluate(lv=lv)

    def get_junc_len_dir(self, variation: str, junc_line):
        '''
        Return the length and direction of a junction defined by a line

        Args:
            variation (str): simulation variation
            junc_line (str): polyline object

        Returns:
            jl (float) : junction length
            uj (list of 3 floats): x,y,z coordinates of the unit vector
                 tangent to the junction line
        '''
        #
        lv = self._get_lv(variation)
        u = []
        for coor in ['X', 'Y', 'Z']:
            calc = CalcObject([], self.setup)
            calc = calc.line_tangent_coor(junc_line, coor)
            u.append(calc.evaluate(lv=lv))

        jl = float(np.sqrt(u[0]**2+u[1]**2+u[2]**2))
        uj = [float(u[0]/jl), float(u[1]/jl), float(u[2]/jl)]
        return jl, uj

    def get_Qseam(self, seam, mode, variation, U_H=None):
        r'''
        Caculate the contribution to Q of a seam, by integrating the current in
        the seam with finite conductance: set in the config file
        ref: http://arxiv.org/pdf/1509.01119.pdf
        '''

        if U_H is None:
            U_H = self.calc_energy_magnetic(variation)

        _, freqs_bare_vals = self.get_freqs_bare(variation)
        self.omega = 2*np.pi*freqs_bare_vals[mode]

        lv = self._get_lv(variation)
        Qseam = OrderedDict()
        print(f'Calculating Qseam_{seam} for mode {mode} ({mode}/{self.n_modes-1})')
        # overestimating the loss by taking norm2 of j, rather than jperp**2
        j_2_norm = self.fields.Vector_Jsurf.norm_2()
        int_j_2 = j_2_norm.integrate_line(seam)
        int_j_2_val = int_j_2.evaluate(lv=lv, phase=90)
        yseam = int_j_2_val/U_H/self.omega

        Qseam['Qseam_'+seam+'_' +
              str(mode)] = config.dissipation.gseam/yseam

        print('Qseam_' + seam + '_' + str(mode), '=', str(config.dissipation.gseam/yseam))

        return pd.Series(Qseam)

    def get_Qseam_sweep(self, seam, mode, variation, variable, values, unit, U_H=None, pltresult=True):
        """
        Q due to seam loss.

        values = ['5mm','6mm','7mm']
        ref: http://arxiv.org/pdf/1509.01119.pdf
        """

        if U_H is None:
            U_H = self.calc_energy_(variation)

        self.solutions.set_mode(mode+1, 0)
        self.fields = self.setup.get_fields()
        freqs_bare_dict, freqs_bare_vals = self.get_freqs_bare(variation)
        self.omega = 2*np.pi*freqs_bare_vals[mode]
        print(variation)
        print(type(variation))
        print(ureg(variation))

        lv = self._get_lv(variation)
        Qseamsweep = []
        print('Calculating Qseam_' + seam + ' for mode ' + str(mode) +
              ' (' + str(mode) + '/' + str(self.n_modes-1) + ')')
        for value in values:
            self.design.set_variable(variable, str(value)+unit)

            # overestimating the loss by taking norm2 of j, rather than jperp**2
            j_2_norm = self.fields.Vector_Jsurf.norm_2()
            int_j_2 = j_2_norm.integrate_line(seam)
            int_j_2_val = int_j_2.evaluate(lv=lv, phase=90)
            yseam = int_j_2_val/U_H/self.omega
            Qseamsweep.append(config.dissipation.gseam/yseam)
#        Qseamsweep['Qseam_sweep_'+seam+'_'+str(mode)] = gseam/yseam
            # Cprint 'Qseam_' + seam + '_' + str(mode) + str(' = ') + str(gseam/yseam)

        if pltresult:
            _, ax = plt.subplots()
            ax.plot(values, Qseamsweep)
            ax.set_yscale('log')
            ax.set_xlabel(variable+' ('+unit+')')
            ax.set_ylabel('Q'+'_'+seam)

        return Qseamsweep

    def get_Qdielectric(self, dielectric, mode, variation, U_E=None):
        if U_E is None:
            U_E = self.calc_energy_electric(variation) 
        Qdielectric = OrderedDict()
        print('Calculating Qdielectric_' + dielectric + ' for mode ' +
              str(mode) + ' (' + str(mode) + '/' + str(self.n_modes-1) + ')')

        U_dielectric = self.calc_energy_electric(variation, obj=dielectric)
        p_dielectric = U_dielectric/U_E
        # TODO: Update make p saved sep. and get Q for diff materials, indep. specify in pinfo
        Qdielectric['Qdielectric_'+dielectric+'_' +
                    str(mode)] = 1/(p_dielectric*config.dissipation.tan_delta_sapp)
        print('p_dielectric'+'_'+dielectric+'_' +
              str(mode)+' = ' + str(p_dielectric))
        return pd.Series(Qdielectric)

    def get_Qsurface_all(self, mode, variation, U_E=None):
        '''
        caculate the contribution to Q of a dieletric layer of dirt on all surfaces
        set the dirt thickness and loss tangent in the config file
        ref: http://arxiv.org/pdf/1509.01854.pdf
        '''
        if U_E is None:
            U_E = self.calc_energy_electric(variation)
        lv = self._get_lv(variation)
        Qsurf = OrderedDict()
        print('Calculating Qsurface for mode ' + str(mode) +
              ' (' + str(mode) + '/' + str(self.n_modes-1) + ')')
#        A = self.fields.Mag_E**2
#        A = A.integrate_vol(name='AllObjects')
#        U_surf = A.evaluate(lv=lv)
        calcobject = CalcObject([], self.setup)
        vecE = calcobject.getQty("E")
        A = vecE
        B = vecE.conj()
        A = A.dot(B)
        A = A.real()
        A = A.integrate_surf(name='AllObjects')
        U_surf = A.evaluate(lv=lv)
        U_surf *= config.dissipation.th*epsilon_0*config.dissipation.eps_r
        p_surf = U_surf/U_E
        Qsurf['Qsurf_'+str(mode)] = 1 / \
            (p_surf*config.dissipation.tan_delta_surf)
        print('p_surf'+'_'+str(mode)+' = ' + str(p_surf))
        return pd.Series(Qsurf)

    def calc_Q_external(self, variation, freq_GHz, U_E = None):
        '''
        Calculate the coupling Q of mode m with each port p
        Expected that you have specified the mode before calling this

        Args:
            variation (str): A string identifier of the variation,
            such as '0', '1', ...
        '''
        if U_E is None:
            U_E = self.calc_energy_electric(variation)
        Qp = pd.Series({})

        freq = freq_GHz * 1e9  # freq in Hz
        for port_nm, port in self.pinfo.ports.items():
            I_peak = self.calc_avg_current_J_surf_mag(variation, port['rect'],
                                                      port['line'])
            U_dissip = 0.5 * port['R'] * I_peak**2 * 1 / freq
            p = U_dissip / (U_E/2)  # U_E is 2x the peak electrical energy
            kappa = p * freq
            Q = 2 * np.pi * freq / kappa
            Qp['Q_' + port_nm] = Q

        return Qp

    def calc_p_junction(self, variation, U_H, U_E, Ljs, Cjs):
        '''
        For a single specific mode.
        Expected that you have specified the mode before calling this, `self.set_mode(num)`

        Expected to precalc U_H and U_E for mode, will retunr pandas pd.Series object
            junc_rect = ['junc_rect1', 'junc_rect2'] name of junc rectangles to integrate H over
            junc_len  = [0.0001]   specify in SI units; i.e., meters
            LJs       = [8e-09, 8e-09] SI units
            calc_sign = ['junc_line1', 'junc_line2']

        WARNING: Cjs is experimental.

        This function assumes there are no lumped capacitors in model.

        Args:
            variation (str): A string identifier of the variation,
            such as '0', '1', ...

        Note:
        --------------
            U_E and U_H are the total peak energy. (NOT twice as in U_ and U_H other places)


        Potential errors:  If you dont have a line or rect by the right name you will prob
        get an erorr o the type:
        com_error: (-2147352567, 'Exception occurred.', (0, None, None, None, 0, -2147024365), None)
        '''

        # ------------------------------------------------------------
        # Calcualte all peak voltage and currents for all junctions in a given mode
        method = self.pinfo.options.method_calc_P_mj

        I_peak_ = {}
        V_peak_ = {}
        Sj = pd.Series({})
        for j_name, j_props in self.pinfo.junctions.items():
            logger.debug(f'Calculating participations for {(j_name, j_props)}')
            Lj = Ljs[j_name]
            Cj = Cjs[j_name]
            line_name = j_props['line']

            if method == 'J_surf_mag':  # old method

                _I_peak_1 = self.calc_avg_current_J_surf_mag(
                    variation, j_props['rect'], line_name)
                # could also use this to back out the V_peak using the impedences as in the line
                # below for now, keep both methods

                _I_peak_2, _V_peak_2, _ = self.calc_current_using_line_voltage(
                    variation, line_name, Lj, Cj)

                logger.debug(
                    f'Differnece in I_Peak calculation ala the two methods: {(_I_peak_1,_I_peak_2)}')

                V_peak = _V_peak_2  # make sure this is signed
                I_peak = _I_peak_1

            elif method == 'line_voltage':  # new preffered method

                I_peak, V_peak, _ = self.calc_current_using_line_voltage(
                    variation, line_name, Lj, Cj)

            else:
                raise NotImplementedError('Other calculation methods\
                    (self.pinfo.options.method_calc_P_mj) are possible but not implemented here. ')

            # save results
            I_peak_[j_name] = I_peak
            V_peak_[j_name] = V_peak
            Sj['s_' + j_name] = _Smj = 1 if V_peak > 0 else - 1

            # REPORT prelimnary
            pmj_ind = 0.5*Ljs[j_name] * I_peak**2 / U_E
            pmj_cap = 0.5*Cjs[j_name] * V_peak**2 / U_E
            #print('\tpmj_ind=',pmj_ind, Ljs[j_name], U_E)

            self.I_peak = I_peak
            self.V_peak = V_peak
            self.Ljs = Ljs
            self.Cjs = Cjs
            print(
                f'\t{j_name:<15} {pmj_ind:>8.6g}{("(+)"if _Smj else "(-)"):>5s}        {pmj_cap:>8.6g}')
            #print('\tV_peak=', V_peak)

        # ------------------------------------------------------------
        # Calcualte participations from the peak voltage and currents
        #

        # All junction capactive and inductive lumped energies - all peak
        U_J_inds = {j_name: 0.5*Ljs[j_name] * I_peak_[j_name]
                    ** 2 for j_name in self.pinfo.junctions}
        U_J_caps = {j_name: 0.5*Cjs[j_name] * V_peak_[j_name]
                    ** 2 for j_name in self.pinfo.junctions}

        U_tot_ind = U_H + sum(list(U_J_inds.values()))  # total
        U_tot_cap = U_E + sum(list(U_J_caps.values()))

        # what to use for the norm?  U_tot_cap or the mean of  U_tot_ind and  U_tot_cap?
        # i.e., (U_tot_ind + U_tot_cap)/2
        U_norm = U_tot_cap
        U_diff = (U_tot_cap-U_tot_ind)/(U_tot_cap+U_tot_ind)
        print("\t\t"f"(U_tot_cap-U_tot_ind)/mean={U_diff*100:.2f}%")
        if abs(U_diff) > 0.15:
            print('WARNING: This simulation must not have converged well!!!\
                The difference in the total cap and ind energies is larger than 10%.\
                Proceed with caution.')

        Pj = pd.Series(OrderedDict([(j_name, Uj_ind/U_norm)
                                    for j_name, Uj_ind in U_J_inds.items()]))

        PCj = pd.Series(OrderedDict([(j_name, Uj_cap/U_norm)
                                     for j_name, Uj_cap in U_J_caps.items()]))

        # print('\t{:<15} {:>8.6g} {:>5s}'.format(
        #    j_name,
        #    Pj['p_' + j_name],
        #    '+' if Sj['s_' + j_name] > 0 else '-'))

        return Pj, Sj, PCj, pd.Series(I_peak), pd.Series(V_peak), \
            {'U_J_inds': U_J_inds,
             'U_J_caps': U_J_caps,
             'U_H': U_H,
             'U_E': U_E,
             'U_tot_ind': U_tot_ind,
             'U_tot_cap': U_tot_cap,
             'U_norm': U_norm,
             'U_diff': U_diff}

    def get_previously_analyzed(self):
        """
        Return previously analyzed data.

        Does not yet handle data that was previously saved in a filename.
        """
        # TODO: maybe load from data_file
        # Rerun previously analyze variations from load filename
        return self._previously_analyzed

    def get_junctions_L_and_C(self, variation: str):
        """
        Returns a pandas Series with the index being the junction name as specified in the
        project_info.

        The values in the series are numeric and in SI base units, i.e., not nH but Henries,
        and not fF but Farads.

        Args:
            variation (str) : label such as '0' or 'all', in which case return
            pandas table for all variations
        """
        if variation == 'all':
            # for all variations and concat
            raise NotImplementedError()  # TODO
        else:
            Ljs = pd.Series({})
            Cjs = pd.Series({})

            for junc_name, val in self.pinfo.junctions.items():  # junction nickname
                _variables = self._hfss_variables[variation]
                def _parse(name): return ureg.Quantity(
                    _variables['_'+val[name]]).to_base_units().magnitude
                Ljs[junc_name] = _parse('Lj_variable')
                Cjs[junc_name] = 2E-15  # _parse(
                # 'Cj_variable') if 'Cj_variable' in val else 0

        return Ljs, Cjs

    def do_EPR_analysis(self,
                        variations: list = None,
                        modes=None,
                        append_analysis=True):
        """
        Main analysis routine

        Args:
            variation (str): A string identifier of the variation,
            such as '0', '1', ...

        Optional Parameters:
        ------------------------
            variations : list | None
                Example list of variations is ['0', '1']
                A variation is a combination of project/design variables in an optimetric sweep

            modes : list | None
                Modes to analyze
                for example  modes = [0, 2, 3]

            append_analysis (bool) : When we run the ansys anslysis, should we redo any variations
                 that we have already done?

        Ansys Notes:
        ------------------------
            Assumptions:
                Low dissipation (high-Q).
                It is easier to assume no lumped capcitors to simply calculations, but we have
                recently added Cj_variable as a new feature that is begin tested to handle capacitors.

                See the paper.


        Using the results:
        ------------------------
            Load results with epr.QuantumAnalysis class


        Example use:
        ----------------

        .. code-block:: python
            :linenos:

            eprd = epr.DistributedAnalysis(pinfo)
            eprd.do_EPR_analysis(append_analysis=False)
        """
        if not modes is None:
            assert max(modes) < self.n_modes, 'Non-existing mode selected. \n'\
                f'The possible modes are between 0 and {self.n_modes-1}.'
            if len(modes) != len(set(modes)):
                logger.warn(f'Select each mode only once! Fixing...\n'\
                    'modes: {modes} --> {list(set(modes))}')
                modes = list(set(modes))

        # Track the total timing
        self._run_time = time.strftime('%Y%m%d_%H%M%S', time.localtime())

        # Update the latest hfss variation information
        self.update_ansys_info()
        variations = variations or self.variations
        modes = modes or range(self.n_modes)
        self.modes = modes

        self.pinfo.save()

        # Main loop - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # TODO: Move inside of loop to funciton calle self.analyze_variation
        for ii, variation in enumerate(variations):
            print(f'\nVariation {variation}  [{ii+1}/{len(variations)}]')

            # Previously analyzed and we should re analyze
            if append_analysis and variation in self.get_previously_analyzed():
                print_NoNewLine('  previously analyzed ...\n')
                continue

            # QUESTION! should we set the current variaiton, can this save time, set the variables

            # If not, clear the results
            self.results[variation] = Dict()
            self.lv = self._get_lv(variation)
            time.sleep(0.4)

            if self.has_fields() == False:
                logger.error(f" Error: HFSS does not have field solution for variation={ii}.\
                                Skipping this mode in the analysis")
                continue

            try:
                # This should allow us to load the fields only once, and then do the calcualtions
                # faster. The loading of the fields does not happen here, but a tthe firc ClcEval call.
                # This could fail if more varialbes are added after the simulation is compelted.
                self.set_variation(variation)
            except Exception as e:
                print('\tERROR: Could not set the variaiton string.'
                      '\nPossible causes: Did you add a variable after the simulation was already solved? '
                      '\nAttempting to proceed nonetheless, should be just slower ...')

            # use nonframe because old style
            freqs_bare_GHz, Qs_bare = self.get_freqs_bare_pd(
                variation, frame=False)

            # update to the latest
            self._hfss_variables[variation] = pd.Series(
                self.get_variables(variation=variation))

            # Create Ljs and Cjs series for a variation
            Ljs, Cjs = self.get_junctions_L_and_C(variation)

            # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
            # This is crummy now.  use dict
            #result = dict()

            Om = OrderedDict()  # Matrix of angular frequency (of analyzed modes)
            Pm = OrderedDict()  # Participation P matrix
            Sm = OrderedDict()  # Sign          S matrix
            Qm_coupling = OrderedDict()  # Quality factor matrix
            SOL = OrderedDict()          # Other results
            Pm_cap = OrderedDict()
            I_peak = OrderedDict()
            V_peak = OrderedDict()
            ansys_energies = OrderedDict()

            for mode in modes:  # integer of mode number [0,1,2,3,..]

                # Load fields for mode
                self.set_mode(mode)

                # Get HFSS  solved frequencies
                _Om = pd.Series({})
                temp_freq = freqs_bare_GHz[mode]
                _Om['freq_GHz'] = temp_freq  # freq
                Om[mode] = _Om
                print(
                    '\n'f'  \033[1mMode {mode} at {"%.2f" % temp_freq} GHz   [{mode+1}/{self.n_modes}]\033[0m')

                # EPR Hamiltonian calculations
                # Calculation global energies and report

                # Magnetic
                print('    Calculating ℰ_magnetic', end=',')
                try:
                    self.U_H = self.calc_energy_magnetic(variation)
                except Exception as e:
                    tb = sys.exc_info()[2]
                    print("\n\nError:\n", e)
                    raise(Exception(' Did you save the field solutions?\n\
                    Failed during calculation of the total magnetic energy.\
                    This is the first calculation step, and is indicative that there are \
                    no field solutions saved. ').with_traceback(tb))

                # Electric
                print('ℰ_electric')
                self.U_E = self.calc_energy_electric(variation)

                # the unnormed
                sol = pd.Series({'U_H': self.U_H, 'U_E': self.U_E})

                # Fraction - report the peak energy, properly normalized
                # the 2 is from the calcualtion methods
                print(f"""     {'(ℰ_E-ℰ_H)/ℰ_E':>15s} {'ℰ_E':>9s} {'ℰ_H':>9s}
    {100*(self.U_E - self.U_H)/self.U_E:>15.1f}%  {self.U_E/2:>9.4g} {self.U_H/2:>9.4g}\n""")

                # Calcualte EPR for each of the junctions
                print(
                    f'    Calculating junction energy participation ration (EPR)\n\tmethod=`{self.pinfo.options.method_calc_P_mj}`. First estimates:')
                print(
                    f"\t{'junction':<15s} EPR p_{mode}j   sign s_{mode}j    (p_capacitive)")

                Pm[mode], Sm[mode], Pm_cap[mode], I_peak[mode], V_peak[mode], ansys_energies[mode] = self.calc_p_junction(
                    variation, self.U_H/2., self.U_E/2., Ljs, Cjs)

                # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                # EPR Dissipative calculations -- should be a function block below

                # TODO: this should really be passed as argument  to the functions rather than a
                # property of the calss I would say
                self.omega = 2*np.pi*freqs_bare_GHz[mode]

                Qm_coupling[mode] = self.calc_Q_external(variation,
                                                         freqs_bare_GHz[mode],
                                                         self.U_E)

                # get seam Q
                if self.pinfo.dissipative['seams']:
                    for seam in self.pinfo.dissipative['seams']:
                        sol = sol.append(self.get_Qseam(seam, mode, variation, self.U_H))

                # get Q dielectric
                if self.pinfo.dissipative['dielectrics_bulk']:
                    for dielectric in self.pinfo.dissipative['dielectrics_bulk']:
                        sol = sol.append(self.get_Qdielectric(
                            dielectric, mode, variation, self.U_E))

                # get Q surface
                if self.pinfo.dissipative['resistive_surfaces']:
                    if self.pinfo.dissipative['resistive_surfaces'] == 'all':
                        sol = sol.append(
                            self.get_Qsurface_all(mode, variation, self.U_E))
                    else:
                        raise NotImplementedError(
                            "Join the team, by helping contribute this piece of code.")

                if self.pinfo.dissipative['resistive_surfaces'] is not None:
                    raise NotImplementedError(
                        "Join the team, by helping contribute this piece of code.")

                SOL[mode] = sol

            # Save
            self._update_results(variation, Om, Pm, Sm, Qm_coupling, SOL,
                                 freqs_bare_GHz, Qs_bare, Ljs, Cjs,
                                 Pm_cap, I_peak, V_peak,
                                 ansys_energies,
                                 self._hfss_variables[variation])
            self.save()

            self._previously_analyzed.add(variation)

        print('\nANALYSIS DONE. Data saved to:\n\n' +
              str(self.data_filename)+'\n\n')

        return self.data_filename, variations

    def _update_results(self, variation: str, Om, Pm, Sm, Qm_coupling, sols,
                        freqs_bare_GHz, Qs_bare, Ljs, Cjs, Pm_cap, I_peak, V_peak,
                        ansys_energies, _hfss_variables):
        '''
        Save variation
        '''
        # raw, not normalized - DataFrames
        self.results[variation]['Pm'] = pd.DataFrame(Pm).transpose()
        self.results[variation]['Pm_cap'] = pd.DataFrame(Pm_cap).transpose()
        self.results[variation]['Sm'] = pd.DataFrame(Sm).transpose()
        self.results[variation]['Om'] = pd.DataFrame(Om)
        self.results[variation]['sols'] = pd.DataFrame(sols).transpose()
        self.results[variation]['Qm_coupling'] = pd.DataFrame(
            Qm_coupling).transpose()

        self.results[variation]['Ljs'] = Ljs  # pd.Series
        self.results[variation]['Cjs'] = Cjs  # pd.Series
        self.results[variation]['Qs'] = Qs_bare
        self.results[variation]['freqs_hfss_GHz'] = freqs_bare_GHz
        self.results[variation]['hfss_variables'] = _hfss_variables
        self.results[variation]['modes'] = self.modes

        # mostly for debug info
        self.results[variation]['I_peak'] = pd.Series(I_peak)
        self.results[variation]['V_peak'] = pd.Series(V_peak)
        self.results[variation]['ansys_energies'] = ansys_energies  # dict

        self.results[variation]['mesh'] = None
        self.results[variation]['convergence'] = None
        self.results[variation]['convergence_f_pass'] = None

        if self.options.save_mesh_stats:
            self.results[variation]['mesh'] = self.get_mesh_statistics(
                variation)  # dataframe
            self.results[variation]['convergence'] = self.get_convergence(
                variation)
            self.results[variation]['convergence_f_pass'] = self.hfss_report_f_convergence(
                variation, save_csv=False)  # dataframe

    @staticmethod
    def results_variations_on_inside(results: dict):
        """
        Switches the order on result of variations. Reverse dict.

        """
        # TODO: THis need to be changed, wont work in the future with updating result etc.
        # if i want to make a base class

        keys = set()
        variations = list(results.keys())
        # Get all keys
        for variation in variations:
            result = results[variation]
            keys.update(result.keys())

        new_res = dict()
        for key in keys:
            new_res[key] = {variation: results[variation].get(key, None)
                            for variation in variations}

            # Conver to pandas Dataframe if all are pd.Series
            if all(isinstance(new_res[key][variation], pd.Series) for variation in variations):
                # print(key) # Conver these to  datafrme
                # Variations will vecome columns
                new_res[key] = pd.DataFrame(new_res[key])
                new_res[key].columns.name = 'variation'
                # sort_df_col : maybe sort

        return new_res  # dict of keys now

    def save(self, project_info: dict = None):
        """Save results to self.data_filename

        Keyword Arguments:
            project_info {dict} -- [description] (default: {None})
        """

        if project_info is None:
            project_info = self.pinfo.save()

        to_save = dict(
            project_info=project_info,
            results=self.results,
        )

        with open(str(self.data_filename), 'wb') as handle:
            pickle.dump(to_save, handle)  # , protocol=pickle.HIGHEST_PROTOCOL)

    def load(self, filepath=None):
        """Utility function to load reuslts file

        Keyword Arguments:
            filepath {[type]} -- [description] (default: {None})
        """
        filepath = filepath or self.data_filename

        with open(str(filepath), 'rb') as handle:
            loaded = pickle.load(handle)

        return loaded

    def get_mesh_statistics(self, variation='0'):
        '''
        Args:
            variation (str): A string identifier of the variation,
            such as '0', '1', ...

        Returns:
        A pandas dataframe, such as

        .. code-block:: text
            :linenos:

                Name	    Num Tets	Min edge    length	    Max edge length	RMS edge length	Min tet vol	Max tet vol	Mean tet vol	Std Devn (vol)
            0	Region	    909451	    0.000243	0.860488	0.037048	    6.006260e-13	0.037352	0.000029	6.268190e-04
            1	substrate	1490356	    0.000270	0.893770	0.023639	    1.160090e-12	0.031253	0.000007	2.309920e-04

        '''
        variation = self._list_variations[ureg(variation)]
        return self.setup.get_mesh_stats(variation)

    def get_convergence(self, variation='0'):
        '''
        Args:
            variation (str): A string identifier of the variation,
            such as '0', '1', ...

        Returns:
            A pandas DataFrame object

            .. code-block:: text
                :linenos:

                    Solved Elements	Max Delta Freq. % Pass Number
                1   	    128955	        NaN
                2       	167607	        11.745000
                3       	192746	        3.208600
                4       	199244	        1.524000

        '''
        variation = self._list_variations[ureg(variation)]
        df, _ = self.setup.get_convergence(variation)
        return df

    def get_convergence_vs_pass(self, variation='0'):
        '''
        Makes a plot in HFSS that return a pandas dataframe

        Args:
            variation (str): A string identifier of the variation,
            such as '0', '1', ...

        Returns:
            Returns a convergence vs pass number of the eignemode freqs.

            .. code-block:: text
                :linenos:

                    re(Mode(1)) [g]	re(Mode(2)) [g]	re(Mode(3)) [g]
                Pass []
                1	4.643101	4.944204	5.586289
                2	5.114490	5.505828	6.242423
                3	5.278594	5.604426	6.296777

        '''
        return self.hfss_report_f_convergence(variation)

    def set_mode(self, mode_num, phase=0):
        '''
        Set source excitations should be used for fields post processing.
        Counting modes from 0 onward
        '''
        assert self.setup, "ERROR: There is no 'setup' connected. \N{face with medical mask}"

        if mode_num < 0:
            logger.error('Too small a mode number')

        self.solutions.set_mode(mode_num + 1, phase)

        if self.has_fields() == False:
            logger.warning(f" Error: HFSS does not have field solution for variation={mode_num}.\
                                    Skipping this mode in the analysis \N{face with medical mask}")

        self.fields = self.setup.get_fields()

    def has_fields(self, variation: str = None):
        '''
        Determine if fields exist for a particular solution.
        Just calls `self.solutions.has_fields(variaiton_string)`

        variation (str | None) : String of variaiton label, such as '0' or '1'
            If None, gets the nominal variation
        '''
        if self.solutions:
            #print('variation=', variation)
            variaiton_string = self.get_variation_string(variation)
            return self.solutions.has_fields(variaiton_string)
        else:
            return False

    def hfss_report_f_convergence(self, variation='0', save_csv=True):
        '''
        Create a report inside HFSS to plot the converge of freq and style it.

        Saves report to csv file.

        Returns a convergence vs pass number of the eignemode freqs.
        Returns a pandas dataframe:

        .. code-block:: text

                re(Mode(1)) [g]	re(Mode(2)) [g]	re(Mode(3)) [g]
            Pass []
            1	4.643101	4.944204	5.586289
            2	5.114490	5.505828	6.242423
            3	5.278594	5.604426	6.296777

        '''
        # TODO: Move to class for reporter ?
        if not self.setup:
            logger.error('NO SETUP PRESENT - hfss_report_f_convergence.')
            return None

        if not self.design.solution_type == 'Eigenmode':
            return None

        oDesign = self.design
        variation = self._get_lv(variation)
        report = oDesign._reporter

        # Create report
        ycomp = [f"re(Mode({i}))" for i in range(1, 1+self.n_modes)]
        params = ["Pass:=", ["All"]]+variation
        report_name = "Freq. vs. pass"
        if report_name in report.GetAllReportNames():
            report.DeleteReports([report_name])
        self.solutions.create_report(
            report_name, "Pass", ycomp, params, pass_name='AdaptivePass')

        # Properties of lines
        curves = [f"{report_name}:re(Mode({i})):Curve1" for i in range(
            1, 1+self.n_modes)]
        set_property(report, 'Attributes', curves, 'Line Width', 3)
        set_property(report, 'Scaling',
                     f"{report_name}:AxisY1", 'Auto Units', False)
        set_property(report, 'Scaling', f"{report_name}:AxisY1", 'Units', 'g')
        set_property(report, 'Legend',
                     f"{report_name}:Legend", 'Show Solution Name', False)

        if save_csv:  # Save
            try:
                path = Path(self.data_dir)/'hfss_eig_f_convergence.csv'
                report.ExportToFile(report_name, path)
                logger.info(f'Saved convergences to {path}')
                return pd.read_csv(path, index_col=0)
            except Exception as e:
                logger.error(f"Error could not save and export hfss plot to {path}.\
                               Is the plot made in HFSS with the correct name.\
                               Check the HFSS error window. \t Error =  {e}")

        return None

    def hfss_report_full_convergence(self, fig=None, _display=True):
        """Plot a full report of teh convergences of an eigenmode analysis for a
        a given variation. Makes a plot inside hfss too.

        Keyword Arguments:
            fig {matpllitb figure} -- Optional figure (default: {None})
            _display {bool} -- Force display or not. (default: {True})

        Returns:
            [type] -- [description]
        """

        if fig is None:
            fig = plt.figure(figsize=(11, 3.))

        for variation in self.variations:
            fig.clf()

            # Grid spec and axes;    height_ratios=[4, 1], wspace=0.5
            gs = mpl.gridspec.GridSpec(1, 3, width_ratios=[1.2, 1.5, 1])
            axs = [fig.add_subplot(gs[i]) for i in range(3)]

            logger.info(f'Creating report for variation {variation}')
            convergence_t = self.get_convergence(variation=variation)
            convergence_f = self.hfss_report_f_convergence(variation=variation)

            ax0t = axs[1].twinx()
            plot_convergence_f_vspass(axs[0], convergence_f)
            plot_convergence_max_df(axs[1], convergence_t.iloc[:, 1])
            plot_convergence_solved_elem(ax0t, convergence_t.iloc[:, 0])
            plot_convergence_maxdf_vs_sol(axs[2], convergence_t.iloc[:, 1],
                                          convergence_t.iloc[:, 0])

            fig.tight_layout(w_pad=0.1)  # pad=0.0, w_pad=0.1, h_pad=1.0)

            if _display:
                from IPython.display import display
                display(fig)

        return fig

    def quick_plot_frequencies(self, swp_variable='variations', ax=None):
        """
        Quick plot of frequencies from HFSS
        """
        fs = self.get_ansys_frequencies_all(swp_variable)
        ax = ax or plt.gca()
        fs['Freq. (GHz)'].unstack(0).transpose().plot(marker='o', ax=ax)
        ax.set_ylabel('Ansys frequencies (MHz)')
        ax.grid(alpha=0.2)
        return fs
